/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOIAnalysis;


/**
 * This class receives an image and calculates some statistical information about the image.
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class ImageStatistical{

	/**
	 * Array where min (minMax[component][0]) and max (minMax[component][1]) values of each component will be stored.
	 * <p>
	 * All values allowed.
	 */
	double minMax[][] = null;

	/**
	 * Global min (minMax[0]) and max (minMax[1]) values of the image.
	 * <p>
	 * All values allowed.
	 */
	double totalMinMax[] = null;

	/**
	 * Average of each component.
	 * <p>
	 * All values allowed.
	 */
	double average[] = null;

	/**
	 * Average of whole image.
	 * <p>
	 * All values allowed.
	 */
	double totalAverage;

	/**
	 * Center of range for each image component.
	 * <p>
	 * All values allowed.
	 */
	double centerRange[] = null;

	/**
	 * Center of range of whole image.
	 * <p>
	 * All values allowed.
	 */
	double totalCenterRange;

	/**
	 * Energy of each component.
	 * <p>
	 * All values allowed.
	 */
	double energy[] = null;

	/**
	 * Energy of whole image.
	 * <p>
	 * All values allowed.
	 */
	double totalEnergy;


	/**
	 * Constructor that does all the operations to calculate min and max, average and center range of the image.
	 *
	 * @param imageSamples a 3D float array that contains image samples
	 */
	public ImageStatistical(float[][][] imageSamples){
		//Size set
		int zSize = imageSamples.length;
		int ySize = imageSamples[0].length;
		int xSize = imageSamples[0][0].length;

		//Memory allocation
		minMax = new double[zSize][2];
		average = new double[zSize];
		centerRange = new double[zSize];
		totalMinMax = new double[2];
		energy = new double[zSize];
		totalAverage = 0.0;
		totalCenterRange = 0.0;
		totalEnergy = 0.0;
		//Initializations
		totalMinMax[0] = Float.POSITIVE_INFINITY;
		totalMinMax[1] = Float.NEGATIVE_INFINITY;
		for(int z = 0; z < zSize; z++){
			minMax[z][0] = Float.POSITIVE_INFINITY;
			minMax[z][1] = Float.NEGATIVE_INFINITY;
			average[z] = 0.0;
			energy[z] = 0.0;
		}

		//Calculus
		for(int z = 0; z < zSize; z++){
		for(int y = 0; y < ySize; y++){
		for(int x = 0; x < xSize; x++){
			//Min and max
			if(imageSamples[z][y][x] < minMax[z][0]){
				minMax[z][0] = imageSamples[z][y][x];
				if(imageSamples[z][y][x] < totalMinMax[0]){
					totalMinMax[0] = imageSamples[z][y][x];
				}
			}
			if(imageSamples[z][y][x] > minMax[z][1]){
				minMax[z][1] = imageSamples[z][y][x];
				if(imageSamples[z][y][x] > totalMinMax[1]){
					totalMinMax[1] = imageSamples[z][y][x];
				}
			}
			//Average
			average[z] += ((double)imageSamples[z][y][x] / (double)(xSize*ySize));
			//Energy
			energy[z] += ((double)(imageSamples[z][y][x] * imageSamples[z][y][x]) / (double)(xSize*ySize));
		}}}
		//centerRange
		for(int z = 0; z < zSize; z++){
			centerRange[z] = (minMax[z][0] + minMax[z][1]) / 2.0;
		}
		//Totals
		for(int z = 0; z < zSize; z++){
			totalAverage += average[z];
			totalCenterRange += centerRange[z];
			totalEnergy += energy[z];
		}
		totalAverage /= (double)zSize;
		totalCenterRange /= (double)zSize;
		totalEnergy /= (double)zSize;
	}

	/**
	 * Constructor that does all the operations to calculate min and max, average and center range of the image.
	 *
	 * @param imageSamples a 3D int array that contains image samples
	 */
	public ImageStatistical(int[][][] imageSamples){
		//Size set
		int zSize = imageSamples.length;
		int ySize = imageSamples[0].length;
		int xSize = imageSamples[0][0].length;

		//Memory allocation
		minMax = new double[zSize][2];
		average = new double[zSize];
		centerRange = new double[zSize];
		totalMinMax = new double[2];
		energy = new double[zSize];
		totalAverage = 0F;
		totalCenterRange = 0F;
		totalEnergy = 0.0;
		//Initializations
		totalMinMax[0] = Float.POSITIVE_INFINITY;
		totalMinMax[1] = Float.NEGATIVE_INFINITY;
		for(int z = 0; z < zSize; z++){
			minMax[z][0] = Float.POSITIVE_INFINITY;
			minMax[z][1] = Float.NEGATIVE_INFINITY;
			average[z] = 0.0;
			energy[z] = 0.0;
		}

		//Calculus
		for(int z = 0; z < zSize; z++){
		for(int y = 0; y < ySize; y++){
		for(int x = 0; x < xSize; x++){
			//Min and max
			if(imageSamples[z][y][x] < minMax[z][0]){
				minMax[z][0] = imageSamples[z][y][x];
				if(imageSamples[z][y][x] < totalMinMax[0]){
					totalMinMax[0] = imageSamples[z][y][x];
				}
			}
			if(imageSamples[z][y][x] > minMax[z][1]){
				minMax[z][1] = imageSamples[z][y][x];
				if(imageSamples[z][y][x] > totalMinMax[1]){
					totalMinMax[1] = imageSamples[z][y][x];
				}
			}
			//Average
			average[z] += ((double)imageSamples[z][y][x] / (double)(xSize*ySize));
			//Energy
			energy[z] += ((double)(imageSamples[z][y][x] * imageSamples[z][y][x]) / (double)(xSize*ySize));
		}}}
		//centerRange
		for(int z = 0; z < zSize; z++){
			centerRange[z] = (minMax[z][0] + minMax[z][1]) / 2.0;
		}
		//Totals
		for(int z = 0; z < zSize; z++){
			totalAverage += average[z];
			totalCenterRange += centerRange[z];
			totalEnergy += energy[z];
		}
		totalAverage /= (double)zSize;
		totalCenterRange /= (double)zSize;
		totalEnergy /= (double)zSize;
	}

	/**
	 * @return minMax definition in this class
	 */
	public double[][] getMinMax(){
		return(minMax);
	}

	/**
	 * @return totalMinMax definition in this class
	 */
	public double[] getTotalMinMax(){
		return(totalMinMax);
	}

	/**
	 * @return average definition in this class
	 */
	public double[] getAverage(){
		return(average);
	}

	/**
	 * @return totalAverage definition in this class
	 */
	public double getTotalAverage(){
		return(totalAverage);
	}

	/**
	 * @return centerRange definition in this class
	 */
	public double[] getCenterRange(){
		return(centerRange);
	}

	/**
	 * @return totalCenterRange definition in this class
	 */
	public double getTotalCenterRange(){
		return(totalCenterRange);
	}

	/**
	 * @return energy definition in this class
	 */
	public double[] getEnergy(){
		return(energy);
	}

	/**
	 * @return totalEnergy definition in this class
	 */
	public double getTotalEnergy(){
		return(totalEnergy);
	}

}
