/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode.Code;
import BOIException.WarningException;
import BOIStream.ByteStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;


/**
 * This class receives a code-block samples and encodes them using the hybrid algorithm.
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class BPECoder{
	final static int maxSize = 128; //Maximum size of the codeblock (must be power of 2)
	final static int yCell = 8; //Size of cells within the codeblock (must be power of 2)
	final static int xCell = 8;
	int maxCellsInPass = 1; //Number of cells in each coding pass

	short[][] statusMap = new short[maxSize][maxSize];
	/* bits 0-4 --> significant bitplane of the coefficient
	 * bit 5 --> 0 if the coefficient is surronded by non-significant coefficients
	 * bits 6-7 --> 0 if the two horizontal coefficients are still non-significant, or they sign is -1 1
	 * bits 6-7 --> 1 if at least 1 horizontal coefficient is negative
	 * bits 6-7 --> 2 if at least 1 horizontal coefficient is positive
	 * bits 8-9 --> same as above but for the vertical coefficients
	 * bit 10 --> 0 if the coefficient is positive
	 * bit 11 --> free
	 * bits 12-15 --> each bit refers to the existence of the top/bottom/left/right neighbor (useful in the boundaries of the codeblock) - 0 means the neighbor exists
	 */

	int[][] listREFS = new int[maxSize * maxSize][2]; //list of refinement bits

	int ySizeCellsMap = -1; //Current size of the cell map
	int xSizeCellsMap = -1;
	int sizeListCells = -1;
	int firstNonVisited = -1;
	int[][] listCellsSIGS = new int[(maxSize/yCell) * (maxSize/xCell)][3];
	/* List of cells sorted by the number of significant coefficients
	 * Each index means [cell][
	 *   0 original position of the cell (useful to sort only this structure)
	 *   1 significant bitplane of the cell
	 *   2 number of significant coefficients in each cell ]
	 */
	int[][] cellsMapCoordinates = new int[(maxSize/yCell) * (maxSize/xCell)][4]; //Coordinates of the cell within the codeblock (non-sorted list, accessed with listCellsSIGS[X][0])

	int LSBPlane = -1; //Least significant bitplane
	int currCellsInPass = -1; //Number of scanned cells in the current coding segment

	/**
	 * ByteStream generated by the fractional bit plane coding.
	 * <p>
	 * The content must be understood as a ByteStream.
	 */
	ByteStream CDByteStream = null;

	/**
	 * Original block samples (index meaning [y][x]).
	 * <p>
	 * All values allowed.
	 */
	int[][] blockSamples = null;

	/**
	 * Map where the contexts of the coefficients are stored. This map is used to extract the context of coefficient when they have to be encoded. It works in the following way: at the beginning all the contextMap is initialized to 0. When a coefficient is found significant, it updates all its neighbors to the next context status and so on. Thanks to this mechanism, less operations have to be performed when a looking for the context of a coefficient.
	 * <p>
	 * The value means the context of the sample. Valid contexts are:
	 * 0 - 8 when the coefficient is still non-significant
	 * 15 - 17 when the coefficient is significant and it is going to be refined
	 */
	byte[][] contextMap = null;

	/**
	 * Block height.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int ySize;

	/**
	 * Block width.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int xSize;

	/**
	 * Image WT subband.
	 * <p>
	 * Valid values: 0(LL), 1(HL), 2(LH), 3(HH).<br>
	 * LL HL<br>
	 * LH HH
	 */
	int subband;

	/**
	 * Coding gain (needed for RD operations).
	 * <p>
	 * It is commonly the L2norm of the wavelet filter bank.
	 */
	float RDAdjust;

	/**
	 * Most Significance Bit Plane of the codeblock.
	 * <p>
	 * Negative values not allowed.
	 */
	int MSBPlane;

	/**
	 * Most Significance Bit Plane of the subband.
	 * <p>
	 * Negative values not allowed.
	 */
	int subbandMSBPlane;

	/**
	 * Error decrease of each truncation point
	 * <p>
	 * Negative values not allowed.
	 */
	float[] errorsDec = null;

	/**
	 * Number of bytes of each coding pass
	 * <p>
	 * Negative values not allowed.
	 */
	int[] lengths = null;

	/**
	 * Definition in {@link BOIDecode.Decode.BPEDecoder#intProcessing}
	 */
	int intProcessing;

	/**
	 * Variable used to calcuate the number of significance/refinement coefficients encoded in the current coding pass.
	 * <p>
	 * Negative values not allowed.
	 */
	private int numSIGS;
	private int numSIGS_CURRBP;
	private int numREFS;
	private int indexListREFS; //numREFS + numSIGS_CURRBP

	/**
	 * Current bit plane encoded.
	 * <p>
	 * 0 is the Less Significant Bit Plane.
	 */
	private int currentBitPlane;

	/**
	 * Current coding pass encoded.
	 * <p>
	 * 0 is the SPP, 1 is MRP, and 2 is CP.
	 */
	private int currentCodingPass;
	private int currentCodingPassGlobal;

	/**
	 * Current mask of the bit plane.
	 * <p>
	 * The content is 2^currentBitPlane
	 */
	static final private int[] currentBitPlaneMask = {
		(1 << 0), (1 << 1), (1 << 2), (1 << 3), (1 << 4), (1 << 5), (1 << 6), (1 << 7), (1 << 8), (1 << 9), (1 << 10), (1 << 11), (1 << 12), (1 << 13), (1 << 14), (1 << 15), (1 << 16), (1 << 17), (1 << 18)
	};
	static final private int[] currentBitPlaneMaskNull = {
		~(currentBitPlaneMask[0]-1), ~(currentBitPlaneMask[1]-1), ~(currentBitPlaneMask[2]-1), ~(currentBitPlaneMask[3]-1), ~(currentBitPlaneMask[4]-1), ~(currentBitPlaneMask[5]-1), ~(currentBitPlaneMask[6]-1), ~(currentBitPlaneMask[7]-1), ~(currentBitPlaneMask[8]-1), ~(currentBitPlaneMask[9]-1), ~(currentBitPlaneMask[10]-1), ~(currentBitPlaneMask[11]-1), ~(currentBitPlaneMask[12]-1), ~(currentBitPlaneMask[13]-1), ~(currentBitPlaneMask[14]-1), ~(currentBitPlaneMask[15]-1), ~(currentBitPlaneMask[16]-1), ~(currentBitPlaneMask[17]-1), ~(currentBitPlaneMask[18]-1)
	};

	private Coder blockCoder = null;


	public BPECoder(){
		blockCoder = new MQCoder();
	}

	public void swapCodeBlock(int[][] blockSamples, int MSBPlane, int subband, int subbandMSBPlane, float RDAdjust, boolean intProcessing) throws WarningException{
		this.blockSamples = blockSamples;
		this.MSBPlane = MSBPlane;
		this.subband = subband;
		this.subbandMSBPlane = subbandMSBPlane;
		this.RDAdjust = RDAdjust;
		this.intProcessing = intProcessing ? 1 : 0;

		boolean sizeChanged = false;
		if(ySize != blockSamples.length || xSize != blockSamples[0].length){
			this.ySize = blockSamples.length;
			this.xSize = blockSamples[0].length;
			this.ySizeCellsMap = (int)Math.ceil((float)ySize / (float)yCell);
			this.xSizeCellsMap = (int)Math.ceil((float)xSize / (float)xCell);
			this.sizeListCells = ySizeCellsMap * xSizeCellsMap;
			this.contextMap = new byte[ySize][xSize];
			sizeChanged = true;
		}
		int bitPlanes = MSBPlane + 1;

		if(ySize*xSize > 4096){
			maxCellsInPass = 2;
		}

		errorsDec = new float[bitPlanes * 512]; //Ridiculously large, but needed for extreme cases
		lengths = new int[bitPlanes * 512];
		CDByteStream = new ByteStream();
		blockCoder.swapOutputByteStream(CDByteStream);

		//Status map
		for(int x = 0; x < xSize; x++){
		for(int y = 0; y < ySize; y++){
			statusMap[y][x] &= 0xF000; //reset everything except boundaries
			if(blockSamples[y][x] < 0){
				statusMap[y][x] |= 0x0400;
				blockSamples[y][x] = -blockSamples[y][x];
			}
			contextMap[y][x] = 0; //reset context
		}}
		if(sizeChanged){
			for(int x = 0; x < xSize; x++){
			for(int y = 0; y < ySize; y++){
				statusMap[y][x] &= 0x0FFF;
			}}
			statusMap[0][0] |= 0xA000;
			statusMap[ySize-1][0] |= 0x6000;
			statusMap[0][xSize-1] |= 0x9000;
			statusMap[ySize-1][xSize-1] |= 0x5000;
			for(int x = 1; x < xSize-1; x++){
				statusMap[0][x] |= 0x8000;
				statusMap[ySize-1][x] |= 0x4000;
			}
			for(int y = 1; y < ySize-1; y++){
				statusMap[y][0] |= 0x2000;
				statusMap[y][xSize-1] |= 0x1000;
			}

			int yBegin = 0;
			int yEnd = yCell;
			for(int yMap = 0; yMap < ySizeCellsMap; yMap++){
				if(yMap == ySizeCellsMap - 1){
					yEnd = ySize;
				}
				int xBegin = 0;
				int xEnd = xCell;
				for(int xMap = 0; xMap < xSizeCellsMap; xMap++){
					if(xMap == xSizeCellsMap - 1){
						xEnd = xSize;
					}
					int cell = yMap * xSizeCellsMap + xMap;
					cellsMapCoordinates[cell][0] = yBegin;
					cellsMapCoordinates[cell][1] = yEnd;
					cellsMapCoordinates[cell][2] = xBegin;
					cellsMapCoordinates[cell][3] = xEnd;
					xBegin = xEnd;
					xEnd += xCell;
				}
				yBegin = yEnd;
				yEnd += yCell;
			}
		}
		int[] numLSBPlane = new int[bitPlanes];
		for(int cell = 0; cell < sizeListCells; cell++){
			listCellsSIGS[cell][0] = cell;
			listCellsSIGS[cell][2] = 0;

			int maxValue = -1;
			for(int y = cellsMapCoordinates[cell][0]; y < cellsMapCoordinates[cell][1]; y++){
			for(int x = cellsMapCoordinates[cell][2]; x < cellsMapCoordinates[cell][3]; x++){
				if(blockSamples[y][x] > maxValue){
					maxValue = blockSamples[y][x];
				}
			}}
			int bp = (int) (Math.log( (float) maxValue) / Math.log(2.0D));
			bp = bp <= 0 ? 1: bp;
			listCellsSIGS[cell][1] = bp;
			numLSBPlane[bp]++;
		}
		//Set the LSBPlane to the bitplane where most cells become significant
		LSBPlane = 0;
		int cumulative = 0;
		for(int bitPlane = MSBPlane; bitPlane > 0; bitPlane--){
			if(numLSBPlane[LSBPlane] < numLSBPlane[bitPlane]){
				LSBPlane = bitPlane;
			}
		}

		//MQ coder
		blockCoder.restart();
		blockCoder.reset();

		//Internal variables of the engine
		currentBitPlane = this.MSBPlane;
		currentCodingPassGlobal = 0;
		currentCodingPass = 0;
		numSIGS = numREFS = numSIGS_CURRBP = indexListREFS = 0;
		currCellsInPass = 0;
		firstNonVisited = 0;
	}

	public void runAll() throws WarningException{
		//Encode the LSBPlane
		int numBits = (int) (Math.log( (float) MSBPlane) / Math.log(2.0D));
		numBits = numBits <= 0 ? 1: numBits;
		for(int i = numBits; i >= 0; i--){
			boolean bit = (LSBPlane & (1 << i)) != 0;
			blockCoder.codeBit(bit, 18);
		}
		//Encode all bitplanes
		while(currentBitPlane >= 0){
			codeNextPass();
		}
		//Terminate the bytestream
		blockCoder.terminate();
		lengths[currentCodingPassGlobal-1] = CDByteStream.getNumBytes();
		renormalizeLengths();
	}

	public void codeNextPass() throws WarningException{
		switch(currentCodingPass){
		case 0:
			if(significanceCoding()){
				if(currCellsInPass == maxCellsInPass){
					errorsDec[currentCodingPassGlobal] = (float) ((double)numSIGS * CoderLUTS.DDsig[intProcessing][subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
					lengths[currentCodingPassGlobal] = CDByteStream.getNumBytes() + blockCoder.remainingBytes();
					currentCodingPassGlobal++;
					currCellsInPass = 0;
					numSIGS_CURRBP += numSIGS;
					numSIGS = 0;
				}
			}else{
				if(currCellsInPass > 0){
					errorsDec[currentCodingPassGlobal] = (float) ((double)numSIGS * CoderLUTS.DDsig[intProcessing][subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
					lengths[currentCodingPassGlobal] = CDByteStream.getNumBytes() + blockCoder.remainingBytes();
					currentCodingPassGlobal++;
					currCellsInPass = 0;
					numSIGS_CURRBP += numSIGS;
					numSIGS = 0;
				}
				currentCodingPass = 1;
			}
			break;
		case 1:
			if(currentBitPlane < MSBPlane){
				refinementCoding();
				errorsDec[currentCodingPassGlobal] = (float) ((double)numREFS * CoderLUTS.DDref[intProcessing][subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
				lengths[currentCodingPassGlobal] = CDByteStream.getNumBytes() + blockCoder.remainingBytes();
				currentCodingPassGlobal++;
			}
			currentCodingPass = 0;
			numREFS += numSIGS_CURRBP;
			numSIGS_CURRBP = 0;
			firstNonVisited = 0;
			currentBitPlane--;
			break;
		}
	}

	private void renormalizeLengths() throws WarningException{
		int currTMPCodingPassGlobal = currentCodingPassGlobal - 2;
		while(currTMPCodingPassGlobal >= 0){
			if(lengths[currTMPCodingPassGlobal] > lengths[currTMPCodingPassGlobal+1]){
				lengths[currTMPCodingPassGlobal] = lengths[currTMPCodingPassGlobal+1];
			}
			if((lengths[currTMPCodingPassGlobal] > 0) && (CDByteStream.getByte(lengths[currTMPCodingPassGlobal] - 1) == (byte)0xFF)){
				lengths[currTMPCodingPassGlobal]--;
			}
			currTMPCodingPassGlobal--;
		}
	}

	static public int MSBPlane(int[][] blockSamples){
		long maxValue = -1;
		int ySize = blockSamples.length;
		int xSize = blockSamples[0].length;
		for(int y = 0; y < ySize; y++){
		for(int x = 0; x < xSize; x++){
			if(Math.abs(blockSamples[y][x]) > maxValue){
				maxValue = Math.abs(blockSamples[y][x]);
			}
		}}
		int bp = (int) (Math.log( (float) maxValue) / Math.log(2.0D));
		return(bp <= 0 ? 1: bp);
	}

	private void foundSignificant(int x, int y){
		//Sign encode
		int signContext = 0;
		switch(statusMap[y][x] & 0x00C0){
		case 0x0040:
			switch(statusMap[y][x] & 0x0300){
			case 0x0100:
				signContext = -14;
				break;
			case 0x0000:
				signContext = -13;
				break;
			case 0x0200:
				signContext = -12;
				break;
			}
			break;
		case 0x0000:
			switch(statusMap[y][x] & 0x0300){
			case 0x0100:
				signContext = -11;
				break;
			case 0x0000:
				signContext = 10;
				break;
			case 0x0200:
				signContext = 11;
				break;
			}
			break;
		case 0x0080:
			switch(statusMap[y][x] & 0x0300){
			case 0x0100:
				signContext = 12;
				break;
			case 0x0000:
				signContext = 13;
				break;
			case 0x0200:
				signContext = 14;
				break;
			}
			break;
		}
		boolean sign = (((statusMap[y][x] & 0x0400) == 0 ? 1 : -1) * (signContext < 0 ? -1 : 1)) < 0 ? true : false;
		blockCoder.codeBit(sign, Math.abs(signContext));

		//Update CPP flag, and sign context
		boolean signNeg = (statusMap[y][x] & 0x0400) != 0;
		if(y >= 1){
			statusMap[y-1][x] |= 0x0020; //CPP
			switch(statusMap[y-1][x] & 0x0300){ //SIGN
			case 0x0000:
				if(signNeg) statusMap[y-1][x] |= 0x0100;
				else statusMap[y-1][x] |= 0x0200;
				break;
			case 0x0100:
				if(!signNeg) statusMap[y-1][x] &= 0xFCFF;
				break;
			case 0x0200:
				if(signNeg) statusMap[y-1][x] &= 0xFCFF;
				break;
			}
			if(x >= 1){
				statusMap[y-1][x-1] |= 0x0020;
			}
			//if(x+1 < xSize){
			if((statusMap[y][x] & 0x1000) == 0){
				statusMap[y-1][x+1] |= 0x0020;
			}
		}
		//if(y+1 < ySize){
		if((statusMap[y][x] & 0x4000) == 0){
			statusMap[y+1][x] |= 0x0020;
			switch(statusMap[y+1][x] & 0x0300){
			case 0x0000:
				if(signNeg) statusMap[y+1][x] |= 0x0100;
				else statusMap[y+1][x] |= 0x0200;
			break;
			case 0x0100:
				if(!signNeg) statusMap[y+1][x] &= 0xFCFF;
			break;
			case 0x0200:
				if(signNeg) statusMap[y+1][x] &= 0xFCFF;
			break;
			}
			if(x >= 1){
				statusMap[y+1][x-1] |= 0x0020;
			}
			//if(x+1 < xSize){
			if((statusMap[y][x] & 0x1000) == 0){
				statusMap[y+1][x+1] |= 0x0020;
			}
		}
		if(x >= 1){
			statusMap[y][x-1] |= 0x0020;
			switch(statusMap[y][x-1] & 0x00C0){
			case 0x0000:
				if(signNeg) statusMap[y][x-1] |= 0x0040;
				else statusMap[y][x-1] |= 0x0080;
			break;
			case 0x0040:
				if(!signNeg) statusMap[y][x-1] &= 0xFF3F;
			break;
			case 0x0080:
				if(signNeg) statusMap[y][x-1] &= 0xFF3F;
			break;
			}
		}
		//if(x+1 < xSize){
		if((statusMap[y][x] & 0x1000) == 0){
			statusMap[y][x+1] |= 0x0020;
			switch(statusMap[y][x+1] & 0x00C0){
			case 0x0000:
				if(signNeg) statusMap[y][x+1] |= 0x0040;
				else statusMap[y][x+1] |= 0x0080;
			break;
			case 0x0040:
				if(!signNeg) statusMap[y][x+1] &= 0xFF3F;
			break;
			case 0x0080:
				if(signNeg) statusMap[y][x+1] &= 0xFF3F;
			break;
			}
		}

		//update currentBP status map
		statusMap[y][x] |= currentBitPlane;

		//update list of refinements
		listREFS[indexListREFS][0] = x;
		listREFS[indexListREFS++][1] = y;

		//increase numSIGS
		numSIGS++;
	}

	private boolean significanceCoding(){
		if(firstNonVisited < sizeListCells){
			boolean scanCell = true;
			if((currentBitPlane > LSBPlane) && (currentBitPlane >= listCellsSIGS[firstNonVisited][1])){
				scanCell = listCellsSIGS[firstNonVisited][1] == currentBitPlane;
				blockCoder.codeBit(scanCell, 9);
			}
			if(scanCell){
				currCellsInPass++;
				int correspondingCell = listCellsSIGS[firstNonVisited][0];
				boolean nullCell = listCellsSIGS[firstNonVisited][1] < currentBitPlane;
				if(!nullCell){
					if(subband == 2){
						for(int y = cellsMapCoordinates[correspondingCell][0]; y < cellsMapCoordinates[correspondingCell][1]; y++){
						for(int x = cellsMapCoordinates[correspondingCell][2]; x < cellsMapCoordinates[correspondingCell][3]; x++){
							if((statusMap[y][x] & 0x001F) == 0){
								boolean significant = (blockSamples[y][x] & currentBitPlaneMask[currentBitPlane]) != 0;
								blockCoder.codeBit(significant, contextMap[y][x]);
								if(significant){
									foundSignificant(x, y);
									updateMap(x, y);
									listCellsSIGS[firstNonVisited][2]++;
								}
							}
						}}
					}else{
						for(int x = cellsMapCoordinates[correspondingCell][2]; x < cellsMapCoordinates[correspondingCell][3]; x++){
						for(int y = cellsMapCoordinates[correspondingCell][0]; y < cellsMapCoordinates[correspondingCell][1]; y++){
							if((statusMap[y][x] & 0x001F) == 0){
								boolean significant = (blockSamples[y][x] & currentBitPlaneMask[currentBitPlane]) != 0;
								blockCoder.codeBit(significant, contextMap[y][x]);
								if(significant){
									foundSignificant(x, y);
									updateMap(x, y);
									listCellsSIGS[firstNonVisited][2]++;
								}
							}
						}}
					}
				}else{ //Null Cell
					if(subband == 2){
						for(int y = cellsMapCoordinates[correspondingCell][0]; y < cellsMapCoordinates[correspondingCell][1]; y++){
						for(int x = cellsMapCoordinates[correspondingCell][2]; x < cellsMapCoordinates[correspondingCell][3]; x++){
							blockCoder.codeBit(false, contextMap[y][x]);
						}}
					}else{
						for(int x = cellsMapCoordinates[correspondingCell][2]; x < cellsMapCoordinates[correspondingCell][3]; x++){
						for(int y = cellsMapCoordinates[correspondingCell][0]; y < cellsMapCoordinates[correspondingCell][1]; y++){
							blockCoder.codeBit(false, contextMap[y][x]);
						}}
					}
				}
			}
			sortCell(firstNonVisited);
			firstNonVisited++;
			return(true);
		}else{
			return(false);
		}
	}

	private void refinementCoding(){
		for(int coef = numREFS-1; coef >= 0; coef--){
			boolean refinement = (blockSamples[listREFS[coef][1]][listREFS[coef][0]] & currentBitPlaneMask[currentBitPlane]) != 0;
			blockCoder.codeBit(refinement, contextMap[listREFS[coef][1]][listREFS[coef][0]]);
		}
	}

	private void sortCell(int cell){
		int newPosition = cell - 1;
		while((newPosition >= 0) && (listCellsSIGS[newPosition][2] < listCellsSIGS[cell][2])){
			newPosition--;
		}
		newPosition++;
		if(newPosition < cell){
			int[] tmp = listCellsSIGS[cell];
			System.arraycopy(listCellsSIGS, newPosition, listCellsSIGS, newPosition + 1, cell - newPosition);
			listCellsSIGS[newPosition] = tmp;
		}
	}

	/**
	 * @return errors definition in {@link #errorsDec}
	 */
	public float[] getErrorsDec(){
		return(errorsDec);
	}

	/**
	 * @return lengths definition in {@link #lengths}
	 */
	public int[] getLengths(){
		return(lengths);
	}

	/**
	 * @return CDByteStream definition in {@link #CDByteStream}
	 */
	public ByteStream getByteStream(){
		return(CDByteStream);
	}

	/**
	 * Turn neighbors map samples of (x,y) from 0 to 1 if they can be visited in the current scan.
	 *
	 * @param x x position of the sample
	 * @param y y position of the sample
	 * @param CP indicates if the updating is performed by the Cleanup Pass (true, otherwise false)
	*/
	private void updateMap(int x, int y){
		//Current coefficient must be refined
		contextMap[y][x] = 17;

		//Context of the neighbors
		switch(subband){
		case 0: //SUBBAND LL
		case 2: //SUBBAND LH
			if(x-1 >= 0){
				//LEFT
				switch(contextMap[y][x-1]){
				case 0:
					contextMap[y][x-1] = (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y][x-1] = (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y][x-1] = (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y][x-1] = (byte) 8;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(contextMap[y-1][x-1]){
					case 0:
						contextMap[y-1][x-1] = (byte) 1;
						break;
					case 1:
						contextMap[y-1][x-1] = (byte) 2;
						break;
					case 5:
						contextMap[y-1][x-1] = (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(contextMap[y+1][x-1]){
					case 0:
						contextMap[y+1][x-1] = (byte) 1;
						break;
					case 1:
						contextMap[y+1][x-1] = (byte) 2;
						break;
					case 5:
						contextMap[y+1][x-1] = (byte) 6;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(contextMap[y][x+1]){
				case 0:
					contextMap[y][x+1] = (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y][x+1] = (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y][x+1] = (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y][x+1] = (byte) 8;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(contextMap[y-1][x+1]){
					case 0:
						contextMap[y-1][x+1] = (byte) 1;
						break;
					case 1:
						contextMap[y-1][x+1] = (byte) 2;
						break;
					case 5:
						contextMap[y-1][x+1] = (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(contextMap[y+1][x+1]){
					case 0:
						contextMap[y+1][x+1] = (byte) 1;
						break;
					case 1:
						contextMap[y+1][x+1] = (byte) 2;
						break;
					case 5:
						contextMap[y+1][x+1] = (byte) 6;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(contextMap[y+1][x]){
				case 0:
					contextMap[y+1][x] = (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y+1][x] = (byte) 3;
					break;
				case 3:
					contextMap[y+1][x] = (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y+1][x] = (byte) 7;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(contextMap[y-1][x]){
				case 0:
					contextMap[y-1][x] = (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y-1][x] = (byte) 3;
					break;
				case 3:
					contextMap[y-1][x] = (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y-1][x] = (byte) 7;
					break;
				}
			}
			break;
		case 1: //SUBBAND HL
			if(x-1 >= 0){
				//LEFT
				switch(contextMap[y][x-1]){
				case 0:
					contextMap[y][x-1] = (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y][x-1] = (byte) 3;
					break;
				case 3:
					contextMap[y][x-1] = (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y][x-1] = (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(contextMap[y-1][x-1]){
					case 0:
						contextMap[y-1][x-1] = (byte) 1;
						break;
					case 1:
						contextMap[y-1][x-1] = (byte) 2;
						break;
					case 5:
						contextMap[y-1][x-1] = (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(contextMap[y+1][x-1]){
					case 0:
						contextMap[y+1][x-1] = (byte) 1;
						break;
					case 1:
						contextMap[y+1][x-1] = (byte) 2;
						break;
					case 5:
						contextMap[y+1][x-1] = (byte) 6;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(contextMap[y][x+1]){
				case 0:
					contextMap[y][x+1] = (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y][x+1] = (byte) 3;
					break;
				case 3:
					contextMap[y][x+1] = (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y][x+1] = (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(contextMap[y-1][x+1]){
					case 0:
						contextMap[y-1][x+1] = (byte) 1;
						break;
					case 1:
						contextMap[y-1][x+1] = (byte) 2;
						break;
					case 5:
						contextMap[y-1][x+1] = (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(contextMap[y+1][x+1]){
					case 0:
						contextMap[y+1][x+1] = (byte) 1;
						break;
					case 1:
						contextMap[y+1][x+1] = (byte) 2;
						break;
					case 5:
						contextMap[y+1][x+1] = (byte) 6;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(contextMap[y+1][x]){
				case 0:
					contextMap[y+1][x] = (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y+1][x] = (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y+1][x] = (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y+1][x] = (byte) 8;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(contextMap[y-1][x]){
				case 0:
					contextMap[y-1][x] = (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y-1][x] = (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y-1][x] = (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y-1][x] = (byte) 8;
					break;
				}
			}
			break;
		case 3: //SUBBAND HH
			if(x-1 >= 0){
				//LEFT
				switch(contextMap[y][x-1]){
				case 0:
					contextMap[y][x-1] = (byte) 1;
					break;
				case 1:
					contextMap[y][x-1] = (byte) 2;
					break;
				case 3:
					contextMap[y][x-1] = (byte) 4;
					break;
				case 4:
					contextMap[y][x-1] = (byte) 5;
					break;
				case 6:
					contextMap[y][x-1] = (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(contextMap[y-1][x-1]){
					case 0:
						contextMap[y-1][x-1] = (byte) 3;
						break;
					case 1:
						contextMap[y-1][x-1] = (byte) 4;
						break;
					case 2:
						contextMap[y-1][x-1] = (byte) 5;
						break;
					case 3:
						contextMap[y-1][x-1] = (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y-1][x-1] = (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y-1][x-1] = (byte) 8;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(contextMap[y+1][x-1]){
					case 0:
						contextMap[y+1][x-1] = (byte) 3;
						break;
					case 1:
						contextMap[y+1][x-1] = (byte) 4;
						break;
					case 2:
						contextMap[y+1][x-1] = (byte) 5;
						break;
					case 3:
						contextMap[y+1][x-1] = (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y+1][x-1] = (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y+1][x-1] = (byte) 8;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(contextMap[y][x+1]){
				case 0:
					contextMap[y][x+1] = (byte) 1;
					break;
				case 1:
					contextMap[y][x+1] = (byte) 2;
					break;
				case 3:
					contextMap[y][x+1] = (byte) 4;
					break;
				case 4:
					contextMap[y][x+1] = (byte) 5;
					break;
				case 6:
					contextMap[y][x+1] = (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(contextMap[y-1][x+1]){
					case 0:
						contextMap[y-1][x+1] = (byte) 3;
						break;
					case 1:
						contextMap[y-1][x+1] = (byte) 4;
						break;
					case 2:
						contextMap[y-1][x+1] = (byte) 5;
						break;
					case 3:
						contextMap[y-1][x+1] = (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y-1][x+1] = (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y-1][x+1] = (byte) 8;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(contextMap[y+1][x+1]){
					case 0:
						contextMap[y+1][x+1] = (byte) 3;
						break;
					case 1:
						contextMap[y+1][x+1] = (byte) 4;
						break;
					case 2:
						contextMap[y+1][x+1] = (byte) 5;
					break;
					case 3:
						contextMap[y+1][x+1] = (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y+1][x+1] = (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y+1][x+1] = (byte) 8;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(contextMap[y+1][x]){
				case 0:
					contextMap[y+1][x] = (byte) 1;
					break;
				case 1:
					contextMap[y+1][x] = (byte) 2;
					break;
				case 3:
					contextMap[y+1][x] = (byte) 4;
					break;
				case 4:
					contextMap[y+1][x] = (byte) 5;
					break;
				case 6:
					contextMap[y+1][x] = (byte) 7;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(contextMap[y-1][x]){
				case 0:
					contextMap[y-1][x] = (byte) 1;
					break;
				case 1:
					contextMap[y-1][x] = (byte) 2;
					break;
				case 3:
					contextMap[y-1][x] = (byte) 4;
					break;
				case 4:
					contextMap[y-1][x] = (byte) 5;
					break;
				case 6:
					contextMap[y-1][x] = (byte) 7;
					break;
				}
			}
			break;
		}
	}
}