/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode.Code;
import BOIStream.ByteStream;


/**
 * This class encodes bits generating a compatible JPEG2000 ByteStream, i.e. a 0xFF is followed by a byte in the range 0x00-0x7F. It could be used by the bypass mode of the fractional bit plane encoding of JPEG2000 or in other situations, for example when encoding packet headers. Usage example:<br>
 * &nbsp; construct<br>
 * &nbsp; codeBit<br>
 * &nbsp; codeBit<br>
 * &nbsp; terminate<br>
 * &nbsp; swapOutputByteStream<br>
 * &nbsp; restart<br>
 * &nbsp; codeBit<br>
 * &nbsp; codeBit<br>
 * &nbsp; ...<br>
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class RawCoder implements BOICode.Code.Coder{

	/**
	 * ByteStream where the generated bytes are flushed
	 *<p>
	 * The ByteStream should be a JPEG2000 compliant codestream.
	 */
	ByteStream outputByteStream;

	/**
	 * Number of bits to transfer (down counter).
	 * <p>
	 * Usually it is initialized to 8, then count down until 0.
	 */
	private int t;

	/**
	 * Byte to transfer.
	 * <p>
	 * Byte flushed to the outputByteStream.
	 */
	private int Tr;


	/**
	 * Constructor that initializes the internal registers of the raw coder.
	 */
	public RawCoder(){
		//Initialize internal registers
		restart();
	}

	/**
	 * Constructor that initializes the internal registers and outputByteStream.
	 *
	 * @param outputByteStream ByteStream where the bytes are flushed
	 */
	public RawCoder(ByteStream outputByteStream){
		//Copy the destination ByteStream
		swapOutputByteStream(outputByteStream);
		//Initialize internal registers
		restart();
	}

	/**
	 * Code a bit.
	 *
	 * @param bit bit to encode
	 */
	public void codeBit(boolean bit){
		t--;
		if(bit){
			Tr |= (1 << t);
		}
		if(t == 0){
			outputByteStream.addByte((byte) Tr);
			if(Tr == 0xFF){
				t = 7;
			}else{
				t = 8;
			}
			Tr = 0;
		}
	}

	/**
	 * Code a bit using some context. The context is not used (this function is only for compatibility).
	 *
	 * @param bit bit to encode
	 * @param context context the bit - not used
	 */
	public void codeBit(boolean bit, int context){
		codeBit(bit);
	}

	public void codeBit(boolean bit, short state){
		codeBit(bit);
	}

	/**
	 * Swaps the current output ByteStream. Before to call this function you should terminate the last ByteStream with the function terminate. Before coding with the new ByteStream you should call the restart function.
	 *
	 * @param outputByteStream destination ByteStream
	 */
	public void swapOutputByteStream(ByteStream outputByteStream){
		this.outputByteStream = outputByteStream;
	}

	/**
	 * Compute the number of bytes belonging to the currently encoded data needed to flush the internal registers. This function is useful to determine potential truncation points for coding passes.
	 *
	 * @return number of bytes
	 */
	public int remainingBytes(){
		//Easy MQ codeword termination
		return((int)Math.ceil((float)t / 8f));
	}


	/**
	 * Finishes the current outputByteStream.
	 */
	public void terminate(){
		if(t < 8){
			outputByteStream.addByte((byte) Tr);
		}
	}

	/**
	 * Restart the value of the internal variables Tr, t, ... of the raw coder.
	 */
	public void restart(){
		Tr = 0;
		t = 8;
	}

	/**
	 * Compatibility with the Coder.
	 */
	public void reset(){}

}
