/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode;
import BOIException.ErrorException;
import BOIException.ParameterException;
import BOIException.WarningException;
import BOIStream.ByteStream;
import BOIFile.LoadFile;
import BOICode.Transform.ColourTransform;
import BOICode.Transform.LevelShift;
import BOICode.Transform.Quantization;
import BOICode.Transform.WaveletTransform;
import BOICode.Code.BlockCode;
import BOICode.Code.BlockDivision;
import BOICode.RateDistortion.LayerCalculation;
import BOICode.RateDistortion.BlockConvexHull;
import BOICode.FileWrite.FileWrite;
import BOICode.FileWrite.JP2Heading;
import BOICode.FileWrite.JPCHeading;
import BOICode.FileWrite.JPKHeading;
import BOICode.FileWrite.PacketHeading;
import BOICode.FileWrite.PrecinctBuild;

import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.StringTokenizer;


/**
 * Main class of BOICoder application. It receives all parameters, checks its validity and runs BOI coder.
 * Usage example:<br>
 * &nbsp; construct<br>
 * &nbsp; setParameters<br>
 * &nbsp; run<br>
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class Coder{

	int[] multipleFiles = null;

	/**
	 * Original image samples (index meaning [z][y][x]).
	 * <p>
	 * All values allowed.
	 */
	float[][][] imageSamplesFloat = null;

	/**
	 * Array that contains the type of each image component.
	 * <p>
	 * Only primitive type value allowed (p.e.: Byte.TYPE, Integer.TYPE, ...).
	 */
	Class[] cTypes = null;

	/**
	 * Indicates if the three first components of the image corresponds to a RGB model.
	 * <p>
	 * True if three first components are RGB, false otherwise.
	 */
	boolean RGBComponents;

	/**
	 * Number of image components.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int zSize;

	/**
	 * Image height.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int ySize;

	/**
	 * Image width.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int xSize;

	/**
	 * Output file name of BOI bitstream.
	 * <p>
	 * Valid file name (with path) and without extension (it's generated automatically using CFileType).
	 */
	String COutFile;

	/**
	 * Header type for the output file:<br>
	 *  <ul>
	 *    <li> 0- NOH Empty headings
	 *    <li> 1- JPC JPEG2000 codestream header
	 *    <li> 2- JP2 JPEG2000 codestream header with JP2 box
	 *    <li> 3- JPK BOI JPEG2000 codestream header
	 *  </ul>
	 */
	int CFileHeader;

	/**
	 * Progression order for the output file:<br>
	 *  <ul>
	 *    <li> 0- LRCP Layer-Resolution-Component-Position
	 *    <li> 1- RLCP Resolution-Layer-Component-Position
	 *    <li> 2- RPCL Resolution-Position-Component-Layer
	 *    <li> 3- PCRL Position-Component-Resolution-Layer
	 *    <li> 4- CPRL Component-Position-Resolution-Layer
	 *  </ul>
	 */
	int CFileProgression;

	/**
	 * Definition in {@link BOICode.FileWrite.FileWrite#FWPacketHeaders}
	 */
	boolean[] FWPacketHeaders = null;

	/**
	 * Vertical offset Tile
	 * <p>
	 * Vertical offset from the origin of the reference grid to the top side of the first tile
	 */
	int YTOsize;

	/**
	 * Hortizontal offset Tile
	 * <p>
	 * Horitzontal offset from the origin of the reference grid to the left side of the first tile
	 */
	int XTOsize;

	/**
	 * Vertical offset.
	 * <p>
	 * Vertical offset from origin of the reference grid to the left side of the image area
	 */
	int YOsize;

	/**
	 * Horitsontal offset.
	 * <p>
	 * Horitzontal offset from origin of the reference grid to the left side of the image area
	 */
	int XOsize;

	/**
	 * Tile height.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int TileYSize;

	/**
	 * Tile width.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int TileXSize;

	/**
	 * Definition in {@link BOICode.Transform.LevelShift#LSSignedComponents}
	 */
	boolean[] LSSignedComponents = null;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTType}
	 */
	int CTType;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTComponents}
	 */
	int[] CTComponents = null;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 */
	int[] WTTypes = null;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 */
	int[] WTLevels = null;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QTypes}
	 */
	int[] QTypes = null;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	int[] QComponentsBits = null;

	/**
	 * Definition in {@link BOICode.Transform.Quantization} (see source code)
	 */
	int QGuardBits;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 */
	int[] BDBlockWidths = null;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 */
	int[] BDBlockHeights = null;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDPrecinctWidths}
	 */
	int[] BDPrecinctWidths = null;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDPrecinctHeights}
	 */
	int[] BDPrecinctHeights = null;

	/**
	 * Definition in {@link BOICode.Code.MQCoder} (see source code)
	 */
	boolean[] MQCFlags = null;

	/**
	 * Definition in {@link BOICode.RateDistortion.LayerCalculation#LCType}
	 */
	int LCType;

	/**
	 * Definition in {@link BOICode.RateDistortion.LayerCalculation#LCTargetNumLayers}
	 */
	int LCTargetNumLayers;

	/**
	 * Definition in {@link BOICode.RateDistortion.LayerCalculation#LCForceSingleLayer}
	 */
	boolean LCForceSingleLayer;

	/**
	 * Definition in {@link BOICode.RateDistortion.LayerCalculation#LCTargetNumBytes}
	 */
	long LCTargetNumBytes;

	/**
	 * Show some information about compression process. Each value is boolean (false means NO show, true means show).
	 * <p>
	 * Each value represents the following:<br>
	 *  <ul>
	 *    <li> 0- Information about time and used memory for each compression stage
	 *    <li> 1- BOI parameters
	 *  </ul>
	 */
	boolean[] CVerbose = null;

	/**
	 * Number of threads employed in the block coding stage.
	 * <p>
	 * Positive values
	 */
	int CThreadNumber = -1;

	//INTERNAL VARIABLES

	/**
	 * To know if parameters are set.
	 * <p>
	 * True indicates that they are set otherwise false.
	 */
	boolean parametersSet = false;

	/**
	 * Used for verbose information (time for stage).
	 * <p>
	 * 0 is initial time.
	 */
	long initStageTime = 0;

	/**
	 * Used for verbose information (total time).
	 * <p>
	 * 0 is initial time.
	 */
	long initTime = 0;


	/**
	 * Constructor of BOICoder. It receives original image and some information about it.
	 *
	 * @param imageSamplesFloat definition in this class
	 * @param cTypes definition in this class
	 * @param RGBComponents definition in this class
	 */
	public Coder(float[][][] imageSamplesFloat, Class[] cTypes, boolean RGBComponents){
		//Image data copy
		this.imageSamplesFloat = imageSamplesFloat;
		this.cTypes = cTypes;
		this.RGBComponents = RGBComponents;

		//Size set
		zSize = imageSamplesFloat.length;
		ySize = imageSamplesFloat[0].length;
		xSize = imageSamplesFloat[0][0].length;
	}

	/**
	 * Set the parameters used to perform the BOI compression and check validity. If a parameter is not initialized, this class initalizes it with default values.
	 *
	 * @param P_COutFile definition in {@link #COutFile}
	 * @param P_CFileHeader definition in {@link #CFileHeader}
	 * @param P_CFileProgression definition in {@link #CFileProgression}
	 * @param P_CTType definition in {@link BOICode.Transform.ColourTransform#CTType}
	 * @param P_CTComponents definition in {@link BOICode.Transform.ColourTransform#CTComponents}
	 * @param P_WTTypes definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 * @param P_WTLevels definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 * @param P_QTypes definition in {@link BOICode.Transform.Quantization#QTypes}
	 * @param P_QGuardBits definition in {@link BOICode.Transform.Quantization}
	 * @param P_BDBlockWidths definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 * @param P_BDBlockHeights definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 * @param P_BDPrecinctWidths definition in {@link BOICode.Code.BlockDivision#BDPrecinctWidths}
	 * @param P_BDPrecinctHeights definition in {@link BOICode.Code.BlockDivision#BDPrecinctHeights}
	 * @param P_LCType definition in {@link BOICode.RateDistortion.LayerCalculation#LCType}
	 * @param P_LCTargetNumLayers definition in {@link BOICode.RateDistortion.LayerCalculation#LCTargetNumLayers}
	 * @param P_LCForceSingleLayer definition in {@link BOICode.RateDistortion.LayerCalculation#LCForceSingleLayer}
	 * @param P_LCTargetNumBytes definition in {@link BOICode.RateDistortion.LayerCalculation#LCTargetNumBytes}
	 * @param P_CVerbose definition in this {@link #CVerbose}
	 * @param P_CThreadNumber definition in this {@link #CThreadNumber}
	 * @throws ParameterException when an invalid or unrecognized exception is detected it will be thrown
	 */
	public void setParameters(
	String    P_COutFile,
	int       P_CFileHeader,
	int       P_CFileProgression,
	int       P_CTType,
	int[]     P_CTComponents,
	int[]     P_WTTypes,
	int[]     P_WTLevels,
	int[]     P_QTypes,
	int       P_QGuardBits,
	int[]     P_BDBlockWidths,
	int[]     P_BDBlockHeights,
	int[]     P_BDPrecinctWidths,
	int[]     P_BDPrecinctHeights,
	int       P_LCType,
	int       P_LCTargetNumLayers,
	int       P_LCForceSingleLayer,
	long      P_LCTargetNumBytes,
	boolean[] P_CVerbose,
	int       P_CThreadNumber,
	int[] P_multipleFiles
	) throws ParameterException{
		parametersSet = true;

		if(P_multipleFiles != null){
			multipleFiles = P_multipleFiles;
		}

		//CANVAS COORDINATE SYSTEM INITIALIZATION (only used as reference information)
		YTOsize = 0;
		XTOsize = 0;
		YOsize = 0;
		XOsize = 0;
		TileYSize = ySize;
		TileXSize = xSize;

		//MQ FLAGS INITIALIZATION (no other configuration is allowed)
		MQCFlags = new boolean[6];
		MQCFlags[0] = false; //BYPASS
		MQCFlags[1] = false; //RESET
		MQCFlags[2] = false; //RESTART
		MQCFlags[3] = false; //CAUSAL
		MQCFlags[4] = false; //ERTERM
		MQCFlags[5] = false; //SEGMARK

		//PACKET HEADER MARKERS (any combination allowed)
		FWPacketHeaders = new boolean[2];
		FWPacketHeaders[0] = false;
		FWPacketHeaders[1] = false;

		//IMAGE COMPONENTS TYPES
		if(cTypes.length != zSize){
			throw new ParameterException("Wrong number of component types.");
		}

		//DATA TYPE USED DETERMINES QComponentsBits and LSSignedComponents
		QComponentsBits = new int[zSize];
		LSSignedComponents = new boolean[zSize];
		for(int z = 0; z < zSize; z++){
			switch(LoadFile.getType(cTypes[z].toString())){
			case 0: //Boolean - boolean (1 byte)
				QComponentsBits[z] = 1;
				LSSignedComponents[z] = false;
				break;
			case 1: //Byte - unsigned int (1 byte)
				QComponentsBits[z] = 8;
				LSSignedComponents[z] = false;
				break;
			case 2: //Char - unsigned int (2 bytes)
				QComponentsBits[z] = 16;
				LSSignedComponents[z] = false;
				break;
			case 3: //Short - signed short (2 bytes)
				QComponentsBits[z] = 16;
				LSSignedComponents[z] = true;
				break;
			case 4: //Int - signed int (4 bytes)
				QComponentsBits[z] = 32;
				LSSignedComponents[z] = true;
				break;
			case 5: //Long - signed long (8 bytes)
				QComponentsBits[z] = 64;
				LSSignedComponents[z] = true;
				break;
			case 6: //Float - float (4 bytes)
				QComponentsBits[z] = 32;
				LSSignedComponents[z] = true;
				break;
			case 7: //Double - double (8 bytes)
				QComponentsBits[z] = 64;
				LSSignedComponents[z] = true;
				break;
			default:
				throw new ParameterException("Unrecognized image component type.");
			}
		}

		//OUTPUT FILE NAME
		if(P_COutFile != null){
			COutFile = P_COutFile;
		}else{
			COutFile = "default";
		}

		//FILE HEADER
		if(P_CFileHeader >= 0){
			if(P_CFileHeader <= 3){
				CFileHeader = P_CFileHeader;
			}else{
				throw new ParameterException("Unrecognized file header type.");
			}
		}else{
			CFileHeader = CoderDefaultValues.CFileHeader;
		}

		//FILE PROGRESSION
		if(P_CFileProgression >= 0){
			if(P_CFileProgression <= 4){
				CFileProgression = P_CFileProgression;
			}else{
				throw new ParameterException("Unrecognized file progression type.");
			}
		}else{
			CFileProgression = CoderDefaultValues.CFileProgression;
		}

		//WAVELET TRANSFORM
		if(P_WTTypes != null){
			if(P_WTTypes.length == zSize){
				for(int z = 0; z < zSize; z++){
					if((P_WTTypes[z] < 0) || (P_WTTypes[z] > 2)){
						throw new ParameterException("\"" + P_WTTypes[z] + "\" is an unrecognized discrete wavelet transform type.");
					}
				}
				WTTypes = P_WTTypes;
			}else{
				if(P_WTTypes.length == 1){
					if((P_WTTypes[0] < 0) || (P_WTTypes[0] > 2)){
						throw new ParameterException("\"" + P_WTTypes[0] + "\" is an unrecognized discrete wavelet transform type.");
					}
					WTTypes = new int[zSize];
					for(int z = 0; z < zSize; z++){
						WTTypes[z] = P_WTTypes[0];
					}
				}else{
					throw new ParameterException("Wrong number of components to do the discrete wavelet transform.");
				}
			}
		}else{
			WTTypes = new int[zSize];
			for(int z = 0; z < zSize; z++){
				WTTypes[z] = CoderDefaultValues.WTType;
			}
		}
		if(P_WTLevels != null){
			if(P_WTLevels.length == zSize){
				for(int z = 0; z < zSize; z++){
					if(P_WTLevels[z] < 0){
						throw new ParameterException("Wavelet transform levels must be positive.");
					}
				}
				WTLevels = P_WTLevels;
			}else{
				if(P_WTLevels.length == 1){
					if(P_WTLevels[0] < 0){
						throw new ParameterException("Wavelet transform levels must be positive.");
					}
					WTLevels = new int[zSize];
					for(int z = 0; z < zSize; z++){
						WTLevels[z] = P_WTLevels[0];
					}
				}else{
					throw new ParameterException("Wrong number of components specifying the number of discrete wavelet transform.");
				}
			}
			//Check WT levels with image size
			for(int z = 0; z < zSize; z++){
				//MaxWTLevels for a subband size (LL) 1x1
				int maxWTLevels = ((int) (Math.log(xSize < ySize ? xSize: ySize) / Math.log(2F)));
				if(WTLevels[z] > maxWTLevels){
					throw new ParameterException("\"" + WTLevels[z] + "\" discrete wavelete transform levels is too much for this image size.");
				}
			}
		}else{
			WTLevels = new int[zSize];
			for(int z = 0; z < zSize; z++){
				//MaxWTLevels for a subband size (LL) of 1x1
				int maxWTLevels = ((int) (Math.log(xSize < ySize ? xSize: ySize) / Math.log(2F)));
				WTLevels[z] = CoderDefaultValues.WTLevels < maxWTLevels ? CoderDefaultValues.WTLevels : maxWTLevels;
			}
		}

		//COLOUR TRANSFORM
		if(P_CTType >= 0){
			if(P_CTType <= 2){
				if(zSize >= 3 || P_CTType == 0){
					CTType = P_CTType;
				}else{
					throw new ParameterException("Colour transform needs a 3 (or more) components image.");
				}
			}else{
				throw new ParameterException("Unrecognized colour transform type.");
			}
		}else{
			if((zSize >= 3) && (RGBComponents)){
				if(WTTypes[0] <= 1){
					CTType = 1;
				}else{
					CTType = 2;
				}
			}else{
				CTType = 0;
			}
		}
		if(P_CTComponents != null){
			if(P_CTComponents.length == 3){
				CTComponents = P_CTComponents;
				for(int z = 0; z < 3; z++){
					if((CTComponents[z] < 0) || (CTComponents[z] >= zSize)){
						throw new ParameterException("Specification of component \"" + CTComponents[z] + "\" in colour transform is wrong (\"-h\" displays help).");
					}
					for(int zCheck = 0; zCheck < zSize; zCheck++){
						if(z != zCheck){
							if(CTComponents[z] == CTComponents[zCheck]){
								throw new ParameterException("All colour transform components must be different.");
							}
						}
					}
				}
			}else{
				throw new ParameterException("Wrong number of components to do the colour transform.");
			}
		}else{
			CTComponents = new int[3];
			for(int z = 0; z < 3; z++){
				CTComponents[z] = z;
			}
		}

		//QUANTIZATION
		if(P_QTypes != null){
			if(P_QTypes.length == zSize){
				for(int z = 0; z < zSize; z++){
					if((P_QTypes[z] < 0) || (P_QTypes[z] > 2)){
						throw new ParameterException("\"" + P_QTypes[z] + "\" is an unrecognized quantization transform type.");
					}
				}
				QTypes = P_QTypes;
			}else{
				if(P_QTypes.length == 1){
					if((P_QTypes[0] < 0) || (P_QTypes[0] > 2)){
						throw new ParameterException("\"" + P_QTypes[0] + "\" is an unrecognized quantization transform type.");
					}
					QTypes = new int[zSize];
					for(int z = 0; z < zSize; z++){
						QTypes[z] = P_QTypes[0];
					}
				}else{
					throw new ParameterException("Wrong number of components to do the quantization transform.");
				}
			}
		}else{
			QTypes = new int[zSize];
			for(int z = 0; z < zSize; z++){
				if(WTTypes[z] == 1){
					QTypes[z] = 0;
				}else{
					QTypes[z] = 2;
				}
			}
		}
		if(P_QGuardBits >= 0){
			if(P_QGuardBits <= 7){
				QGuardBits = P_QGuardBits;
			}else{
				throw new ParameterException("Guard bit number must be between 0 and 7.");
			}
		}else{
			QGuardBits = CoderDefaultValues.QGuardBits;
		}

		//BLOCK DIVISION
		if(P_BDBlockWidths != null){
			if(P_BDBlockWidths.length == zSize){
				for(int z = 0; z < zSize; z++){
					if(P_BDBlockWidths[z] < 2){
						throw new ParameterException("Block width must be equal or greater than 2 for all components.");
					}
				}
				BDBlockWidths = P_BDBlockWidths;
			}else{
				if(P_BDBlockWidths.length == 1){
					if(P_BDBlockWidths[0] < 2){
						throw new ParameterException("Block width must be equal or greater than 2 for all components.");
					}
					BDBlockWidths = new int[zSize];
					for(int z = 0; z < zSize; z++){
						BDBlockWidths[z] = P_BDBlockWidths[0];
					}
				}else{
					throw new ParameterException("Wrong number of components specifying the number of block width.");
				}
			}
		}else{
			BDBlockWidths = new int[zSize];
			for(int z = 0; z < zSize; z++){
				BDBlockWidths[z] = CoderDefaultValues.BDBlockWidth;
			}
		}
		if(P_BDBlockHeights != null){
			if(P_BDBlockHeights.length == zSize){
				for(int z = 0; z < zSize; z++){
					if(P_BDBlockHeights[z] < 2){
						throw new ParameterException("Block height must be equal or greater than 2 for all components.");
					}
				}
				BDBlockHeights = P_BDBlockHeights;
			}else{
				if(P_BDBlockHeights.length == 1){
					if(P_BDBlockHeights[0] < 2){
						throw new ParameterException("Block height must be equal or greater than 2 for all components.");
					}
					BDBlockHeights = new int[zSize];
					for(int z = 0; z < zSize; z++){
						BDBlockHeights[z] = P_BDBlockHeights[0];
					}
				}else{
					throw new ParameterException("Wrong number of components specifying the number of block height.");
				}
			}
		}else{
			BDBlockHeights = new int[zSize];
			for(int z = 0; z < zSize; z++){
				BDBlockHeights[z] = CoderDefaultValues.BDBlockHeight;
			}
		}

		//PRECINCT SIZES
		int maxPrecinctWidth = (int) Math.ceil((Math.log(xSize) / Math.log(2F)));
		if(P_BDPrecinctWidths != null){
			if(P_BDPrecinctWidths.length == zSize){
				for(int z = 0; z < zSize; z++){
					if(P_BDPrecinctWidths[z] <= BDBlockWidths[z]){
						throw new ParameterException("Precinct width must be greater than block width for all components.");
					}
					if(P_BDPrecinctWidths[z] > maxPrecinctWidth){
						throw new ParameterException("Precinct width must be equal or less than image width for all components (max width is " + maxPrecinctWidth + ").");
					}
				}
				BDPrecinctWidths = P_BDPrecinctWidths;
			}else{
				if(P_BDPrecinctWidths.length == 1){
					if(P_BDPrecinctWidths[0] > maxPrecinctWidth){
						throw new ParameterException("Precinct width must be equal or less than image width for all components.");
					}
					for(int z = 0; z < zSize; z++){
						if(P_BDPrecinctWidths[0] <= BDBlockWidths[z]){
							throw new ParameterException("Precinct width must be greater than block width for all components.");
						}
					}
					BDPrecinctWidths = new int[zSize];
					for(int z = 0; z < zSize; z++){
						BDPrecinctWidths[z] = P_BDPrecinctWidths[0];
					}
				}else{
					throw new ParameterException("Wrong number of components specifying the number of precinct width.");
				}
			}
		}else{
			BDPrecinctWidths = new int[zSize];
			for(int z = 0; z < zSize; z++){
				BDPrecinctWidths[z] = maxPrecinctWidth;
			}
		}
		int maxPrecinctHeight = (int) Math.ceil((Math.log(ySize) / Math.log(2F)));
		if(P_BDPrecinctHeights != null){
			if(P_BDPrecinctHeights.length == zSize){
				for(int z = 0; z < zSize; z++){
					if(P_BDPrecinctHeights[z] <= BDBlockHeights[z]){
						throw new ParameterException("Precinct height must be greater than block height for all components.");
					}
					if(P_BDPrecinctHeights[z] > maxPrecinctHeight){
						throw new ParameterException("Precinct height must be equal or less than image height for all components (max height is " + maxPrecinctHeight + ").");
					}
				}
				BDPrecinctHeights = P_BDPrecinctHeights;
			}else{
				if(P_BDPrecinctHeights.length == 1){
					if(P_BDPrecinctHeights[0] > maxPrecinctHeight){
						throw new ParameterException("Precinct height must be equal or less than image height for all components.");
					}
					for(int z = 0; z < zSize; z++){
						if(P_BDPrecinctHeights[0] <= BDBlockHeights[z]){
							throw new ParameterException("Precinct height must be greater than block height for all components.");
						}
					}
					BDPrecinctHeights = new int[zSize];
					for(int z = 0; z < zSize; z++){
						BDPrecinctHeights[z] = P_BDPrecinctHeights[0];
					}
				}else{
					throw new ParameterException("Wrong number of components specifying the number of precinct height.");
				}
			}
		}else{
			BDPrecinctHeights = new int[zSize];
			for(int z = 0; z < zSize; z++){
				BDPrecinctHeights[z] = maxPrecinctHeight;
			}
		}

		//LAYER CALCULATION TYPE
		if(P_LCType >= 0){
			if(P_LCType <= 3){
				LCType = P_LCType;
			}else{
				throw new ParameterException("Unrecognized layer calculation type.");
			}
		}else{
			LCType = CoderDefaultValues.LCType;
		}
		if(P_LCTargetNumBytes >= 0){
			LCTargetNumBytes = P_LCTargetNumBytes;
		}else{
			LCTargetNumBytes = CoderDefaultValues.LCTargetNumBytes;
		}
		if(P_LCTargetNumLayers >= 0){
			LCTargetNumLayers = P_LCTargetNumLayers;
		}else{
			LCTargetNumLayers = CoderDefaultValues.LCTargetNumLayers;
		}
		if((LCType > 0) && ((LCTargetNumBytes > 0) && (LCTargetNumLayers > 0))){
			throw new ParameterException("With layer calculation type " + LCType + " can not specify number of bytes and layers simultaneously (\"-h\" displays help).");
		}
		if(P_LCForceSingleLayer >= 0){
			if(LCType > 0){
				switch(P_LCForceSingleLayer){
				case 0:
					LCForceSingleLayer = false;
					break;
				case 1:
					LCForceSingleLayer = true;
					break;
				default:
					throw new ParameterException("Unrecognized forceSingleLayer type.");
				}
			}else{
				throw new ParameterException("With layerCalculationType " + LCType + " forceSingleLayer cannot be specified (\"-h\" displays help).");
			}
		}else{
			LCForceSingleLayer = CoderDefaultValues.LCForceSingleLayer;
		}

		//VERBOSE INFORMATION
		if(P_CVerbose != null){
			CVerbose = P_CVerbose;
		}else{
			CVerbose = CoderDefaultValues.CVerbose;
		}

		//NUMBER OF THREADS
		if(P_CThreadNumber > 0){
			CThreadNumber = P_CThreadNumber;
		}else{
			CThreadNumber = CoderDefaultValues.CThreadNumber;
		}

	}

	/**
	 * Runs the BOI coder algorithm to compress the image.
	 *
	 * @throws ErrorException when something goes wrong and compression must be stopped
	 */
	public void run() throws ErrorException{
		//If parameters are not set run cannot be executed
		if(!parametersSet){
			throw new ErrorException("Parameters not set.");
		}

		//Image samples in different types
		int[][][]         imageSamplesInt = null;
		int[][][][][][][] imageBlocks = null;
		//Definitions in Transform/Quantization
		int[][][] QExponents = null;
		int[][][] QMantisas = null;
		//Definitions in Code/BlockDivision
		int[][] BDResolutionPrecinctWidths = null;
		int[][] BDResolutionPrecinctHeights = null;
		int[][] BDBlocksPerPrecinctWidths = null;
		int[][] BDBlocksPerPrecinctHeights = null;
		//Definitions in Code/BlockCode
		int[][][][][]            BCMSBPlanes = null;
		float[][][][][][]      BCErrorsDec = null;
		int[][][][][][]        BCLengths = null;
		ByteStream[][][][][]     BCByteStreams = null;
		//Definitions in RateDistortion/BlockConvexHull
		int[][][][][][]   BCHFeasiblePoints = null;
		float[][][][][][] BCHSlopes = null;
		//To know number of bytes and layers generated
		long LCAchievedNumBytes = 0;
		int  LCAchievedNumLayers = 0;
		//Definitions in RateDistortion/LayerCalculation
		int[][][][][][] LCLayers = null;
		//Definitions in FileWrite/PrecinctBuild
		ByteStream[][][][][][][] PBByteStreamsLayers = null;
		int[][][][][][][][]      PBByteStreamsLayersRange = null;
		int[][][][][][][]        PBByteStreamsLayersCodingPasses = null;
		int[][][][][][]          PBMostBitPlanesNull = null;
		int[][][][][][]          PBFirstLayer = null;
		//Definitions in FileWrite/PacketHeading
		ByteStream[][][][] PHHeaders = null;

		showTimeMemory("IMAGE LOAD. STARTING COMPRESSION...");

		//LEVEL SHIFT
		LevelShift ls = new LevelShift(imageSamplesFloat);
		ls.setParameters(1, QComponentsBits, LSSignedComponents);
		imageSamplesFloat = ls.run();
		//Free unused memory
		ls = null;
		//Show statistics
		showTimeMemory("level shift");

		//COLOUR TRANSFORM
		ColourTransform ct = new ColourTransform(imageSamplesFloat);
		ct.setParameters(CTType, CTComponents);
		imageSamplesFloat = ct.run();
		//Free unused memory
		ct = null;
		//Show statistics
		showTimeMemory("colour transform");

		//DISCRETE WAVELET TRANSFORM
		WaveletTransform wt = new WaveletTransform(imageSamplesFloat);
		wt.setParameters(WTTypes, WTLevels);
		imageSamplesFloat = wt.run();
		//Free unused memory
		wt = null;
		//Show statistics
		showTimeMemory("discrete wavelet transform");

		//QUANTIZATION
		Quantization q = new Quantization(imageSamplesFloat);
		q.setParameters(WTLevels, QTypes, QComponentsBits);
		imageSamplesInt = q.run();
		QExponents = q.getExponents();
		QMantisas = q.getMantisas();
		//Free unused memory
		q = null;
		imageSamplesFloat = null;
		//Show statistics
		showTimeMemory("quantization");

		//BLOCK AND PRECINCT DIVISION
		BlockDivision bd = new BlockDivision(imageSamplesInt);
		bd.setParameters(WTLevels, BDBlockWidths, BDBlockHeights, BDPrecinctWidths, BDPrecinctHeights);
		imageBlocks = bd.run();
		BDResolutionPrecinctWidths = bd.getResolutionPrecinctWidths();
		BDResolutionPrecinctHeights = bd.getResolutionPrecinctHeights();
		BDBlocksPerPrecinctWidths = bd.getBlocksPerPrecinctWidths();
		BDBlocksPerPrecinctHeights = bd.getBlocksPerPrecinctHeights();
		//Free unused memory
		bd = null;
		imageSamplesInt = null;
		//Show statistics
		showTimeMemory("block and precinct division");

		//BIT PLANE BLOCK CODE
		BlockCode.initialize(imageBlocks, WTTypes, CThreadNumber);
		BlockCode[] bc = new BlockCode[CThreadNumber];
		for(int thread = 0; thread < CThreadNumber; thread++){
			bc[thread] = new BlockCode(thread);
		}
		Thread[] threads = new Thread[CThreadNumber];
		for(int thread = 0; thread < CThreadNumber; thread++){
			threads[thread] = new Thread(bc[thread]);
			threads[thread].start();
		}
		for(int thread = 0; thread < CThreadNumber; thread++){
			try{
				threads[thread].join();
			}catch(InterruptedException e){
				throw new ErrorException(e.getMessage());
			}
		}
		BCMSBPlanes = BlockCode.getMSBPlanes();
		BCErrorsDec = BlockCode.getErrorsDec();
		BCLengths = BlockCode.getLengths();
		BCByteStreams = BlockCode.getByteStreams();
		//Free unused memory
		bc = null;
		threads = null;
		imageBlocks = null;
		//Show statistics
		showTimeMemory("bit plane block coding + mq coding");

		//BLOCK CONVEX HULL CALCULATION
		BlockConvexHull bch = null;
		bch = new BlockConvexHull(BCErrorsDec, BCLengths);
		bch.setParameters();
		bch.run();
		BCHFeasiblePoints = bch.getFeasiblePoints();
		BCHSlopes = bch.getSlopes();
		//Free unused memory
		bch = null;
		BCErrorsDec = null;
		//Show statistics
		showTimeMemory("rate distortion");


		//GENERATION OF MULTIPLE FILES
		long incTargetBytes = 0;
		int numFile = -1;
		if(multipleFiles != null){
			LCTargetNumBytes = multipleFiles[0];
			incTargetBytes = multipleFiles[1];
			numFile = 1;
		}
		boolean maxFile = false;
		long lastFile = 0;
		do{
		//GENERATION OF MULTIPLE FILES

		//LAYERS CALCULATION
		LayerCalculation lc = new LayerCalculation(BCLengths, BCHFeasiblePoints, BCHSlopes);
		lc.setParameters(LCType, LCTargetNumBytes, LCTargetNumLayers, LCForceSingleLayer, WTTypes, null);
		lc.run();
		LCAchievedNumBytes = lc.getAchievedNumBytes();
		LCAchievedNumLayers = lc.getAchievedNumLayers();
		LCLayers = lc.getLayers();
		//Free unused memory (needed below)
		//lc = null;
		//BCHFeasiblePoints = null;
		//BCHSlopes = null;
		//Show statistics
		showTimeMemory("layer calculation");
		//Free unused memory (needed below)
		//BCLengths = null;

		maxFile = lastFile == LCAchievedNumBytes;
		lastFile = LCAchievedNumBytes;

		//PRECINCT BUILD
		PrecinctBuild pb = null;
		pb = new PrecinctBuild(BCByteStreams, BCLengths, LCAchievedNumLayers, LCLayers, BDBlocksPerPrecinctWidths, BDBlocksPerPrecinctHeights, QExponents, QGuardBits, BCMSBPlanes);
		pb.run();
		PBByteStreamsLayers = pb.getByteStreamsLayers();
		PBByteStreamsLayersRange = pb.getByteStreamsLayersRange();
		PBByteStreamsLayersCodingPasses = pb.getByteStreamsLayersCodingPasses();
		PBMostBitPlanesNull = pb.getMostBitPlanesNull();
		PBFirstLayer = pb.getFirstLayer();
		//Free unused memory
		pb = null;
		//Show statistics
		showTimeMemory("precinct build");

		//PACKET HEADING
		PacketHeading ph = null;
		ph = new PacketHeading(PBByteStreamsLayers, PBByteStreamsLayersRange, PBByteStreamsLayersCodingPasses, PBMostBitPlanesNull, PBFirstLayer);
		PHHeaders = ph.run();
		//Free unused memory
		ph = null;
		//Show statistics
		showTimeMemory("packet heading generation");

		//Generation of quality layers is launched again to better adjust the targeted bytes once the packet headers lengths are known
		if((LCType == 0) && (LCTargetNumBytes > 0)){

			//LAYERS CALCULATION
			LCLayers = null;
			lc.setParameters(LCType, LCTargetNumBytes, LCTargetNumLayers, LCForceSingleLayer, WTTypes, PHHeaders);
			lc.run();
			LCAchievedNumBytes = lc.getAchievedNumBytes();
			LCAchievedNumLayers = lc.getAchievedNumLayers();
			LCLayers = lc.getLayers();
			//Free unused memory
			lc = null;
			//BCHFeasiblePoints = null;
			//BCHSlopes = null;
			//Show statistics
			showTimeMemory("layer calculation");
			//Free unused memory
			//BCLengths = null;

			//PRECINCT BUILD
			pb = null;
			pb = new PrecinctBuild(BCByteStreams, BCLengths, LCAchievedNumLayers, LCLayers, BDBlocksPerPrecinctWidths, BDBlocksPerPrecinctHeights, QExponents, QGuardBits, BCMSBPlanes);
			pb.run();
			PBByteStreamsLayers = pb.getByteStreamsLayers();
			PBByteStreamsLayersRange = pb.getByteStreamsLayersRange();
			PBByteStreamsLayersCodingPasses = pb.getByteStreamsLayersCodingPasses();
			PBMostBitPlanesNull = pb.getMostBitPlanesNull();
			PBFirstLayer = pb.getFirstLayer();
			//Free unused memory
			pb = null;
			//Show statistics
			showTimeMemory("precinct build");

			//PACKET HEADING
			ph = null;
			PHHeaders = null;
			ph = new PacketHeading(PBByteStreamsLayers, PBByteStreamsLayersRange, PBByteStreamsLayersCodingPasses, PBMostBitPlanesNull, PBFirstLayer);
			PHHeaders = ph.run();
			//Free unused memory
			ph = null;
			//Show statistics
			showTimeMemory("packet heading generation");
		}

		//Free unused memory (PRECINCT BUILD)
		/*BCByteStreams = null;
		BCMSBPlanes = null;
		BCLengths = null;
		//Free unused memory (PACKET HEADING)
		PBMostBitPlanesNull = null;
		PBFirstLayer = null;
		PBByteStreamsLayersCodingPasses = null;*/

		//HEADINGS AND FILE WRITE

		//File name
		String[] strHeading          = {"noh", "jpc", "jp2", "jpk"};
		String outFile = "";
		if(numFile > 0){
			outFile = COutFile + "." + numFile  + "." + strHeading[CFileHeader];
		}else{
			outFile = COutFile + "." + strHeading[CFileHeader];
		}
		try{
			//Headings
			ByteStream mainHeading = new ByteStream();
			JPCHeading JPCH = null;
			JP2Heading JP2H = null;
			JPKHeading JPKH = null;
			switch(CFileHeader){
			case 0: //NOH Empty heading
				break;
			case 1: //JPC JPEG2000 codestream header
				JPCH = new JPCHeading(zSize, ySize, xSize, ySize, xSize, LSSignedComponents, CTType, WTTypes, WTLevels, QTypes, QComponentsBits, QGuardBits, QExponents, QMantisas, BDBlockHeights, BDBlockWidths, BDResolutionPrecinctHeights, BDResolutionPrecinctWidths, MQCFlags, LCAchievedNumLayers, CFileProgression, FWPacketHeaders);
				mainHeading = JPCH.run();
				break;
			case 2: //JP2 JPEG2000 codestream header with JP2 box
				ByteStream JPCHeading = new ByteStream();
				JPCH = new JPCHeading(zSize, ySize, xSize, ySize, xSize, LSSignedComponents, CTType, WTTypes, WTLevels, QTypes, QComponentsBits, QGuardBits, QExponents, QMantisas, BDBlockHeights, BDBlockWidths, BDResolutionPrecinctHeights, BDResolutionPrecinctWidths, MQCFlags, LCAchievedNumLayers, CFileProgression, FWPacketHeaders);
				JPCHeading = JPCH.run();

				ByteStream JP2Heading = new ByteStream();
				JP2H = new JP2Heading(zSize, ySize, xSize, LSSignedComponents, CTType, QComponentsBits);
				JP2Heading = JP2H.run();

				mainHeading.addBytes(JP2Heading.getByteStream(), JP2Heading.getNumBytes());
				mainHeading.addBytes(JPCHeading.getByteStream(), JPCHeading.getNumBytes());
				break;
			case 3: //BOI JPK JPEG2000 codestream header
				JPKH = new JPKHeading(zSize, ySize, xSize, LSSignedComponents, CTType, CTComponents, WTTypes, WTLevels, QTypes, QComponentsBits, QExponents, QMantisas, QGuardBits, BDBlockWidths, BDBlockHeights, BDResolutionPrecinctWidths, BDResolutionPrecinctHeights, LCAchievedNumLayers, CFileProgression, FWPacketHeaders);
				mainHeading = JPKH.run();
				break;
			}

			//Write file
			FileWrite FW = null;
			FW = new FileWrite(zSize, ySize, xSize, YTOsize, XTOsize, YOsize, XOsize, TileYSize, TileXSize, BDResolutionPrecinctWidths, BDResolutionPrecinctHeights, WTLevels, LCLayers, BDBlocksPerPrecinctWidths, BDBlocksPerPrecinctHeights, mainHeading, PBByteStreamsLayers, PBByteStreamsLayersRange, PHHeaders, CFileProgression, FWPacketHeaders, outFile);
			FW.run();

			//Free unused memory
			mainHeading = null;
			JPCH = null;
			JP2H = null;
			FW = null;
			//Show statistics
			showTimeMemory("write file (" + outFile + ")");

		}catch(WarningException e){
			throw new ErrorException("The codestream cannot be saved: " + e.getMessage());
		}

		showParameters(QExponents, QMantisas, BDResolutionPrecinctWidths, BDResolutionPrecinctHeights, LCAchievedNumLayers);

		numFile++;
		LCTargetNumBytes += incTargetBytes;
		}while(!maxFile && numFile > 0);

		//Free unused memory
		//QExponents = null;
		//QMantisas = null;
		//BDResolutionPrecinctWidths = null;
		//BDResolutionPrecinctHeights = null;
		//BDBlocksPerPrecinctWidths = null;
		//BDBlocksPerPrecinctHeights = null;
		//BCByteStreams = null;
		//LCLayers = null;
		//PBByteStreamsLayers = null;
		//PBByteStreamsLayersRange = null;
		//PHHeaders = null;
	}

	//////////////////////////////////////////////
	//FUNCTIONS TO SHOW SOME VERBOSE INFORMATION//
	//////////////////////////////////////////////

	/**
	 * Show some time and memory usage statisticals.
	 *
	 * @param stage string that will be displayed
	 */
	void showTimeMemory(String stage){
		if(CVerbose[0]){
			long actualTime = System.currentTimeMillis();
			if(initTime == 0){
				initTime = actualTime;
			}
			if(initStageTime == 0){
				initStageTime = actualTime;
			}

			//print times are not considered
			long totalMemory = Runtime.getRuntime().totalMemory() / 1048576;
			long usedMemory =  (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()) / 1048576;
			//((float) initTime - (float) actualTime)
			String durationStage = Float.toString((actualTime - initStageTime) / 1000F) + "000";
			durationStage = durationStage.substring(0, durationStage.lastIndexOf(".") + 4);
			String duration = Float.toString((actualTime - initTime) / 1000F) + "000";
			duration = duration.substring(0, duration.lastIndexOf(".") + 4);

			System.out.println("STAGE: " + stage);
			System.out.println("  Memory (USED/TOTAL): " + usedMemory + "/" + totalMemory + " MB");
			System.out.println("  Time (USED/TOTAL)  : " + durationStage + "/" + duration + " secs");

			initStageTime = System.currentTimeMillis();
		}
	}

	/**
	 * Show BOI image parameters.
	 *
	 * @param QExponents needed because is a not global variable
	 * @param QMantisas needed because is a not global variable
	 * @param BDResolutionPrecinctWidths needed because is a not global variable
	 * @param BDResolutionPrecinctHeights needed because is a not global variable
	 * @param LCAchievedNumLayers needed because is a not global variable
	 */
	void showParameters(int[][][] QExponents, int[][][] QMantisas, int[][] BDResolutionPrecinctWidths, int[][] BDResolutionPrecinctHeights, int LCAchievedNumLayers){
		if(CVerbose[1]){
			System.out.print("BOI PARAMETERS");

			System.out.print("\n  zSize: " + zSize);
			System.out.print("\n  ySize: " + ySize);
			System.out.print("\n  xSize: " + xSize);

			System.out.print("\n  LSSignedComponents: ");
			for(int z = 0; z < LSSignedComponents.length; z++){
				System.out.print(LSSignedComponents[z] + " ");
			}

			System.out.print("\n  CTType: " + CTType);
			System.out.print("\n  CTComponents: ");
			for(int i = 0; i < CTComponents.length; i++){
				System.out.print(CTComponents[i] + " ");
			}

			System.out.print("\n  WTTypes: ");
			for(int z = 0; z < WTTypes.length; z++){
				System.out.print(WTTypes[z] + " ");
			}
			System.out.print("\n  WTLevels: ");
			for(int z = 0; z < WTLevels.length; z++){
				System.out.print(WTLevels[z] +  " ");
			}

			System.out.print("\n  QTypes: ");
			for(int z = 0; z < QTypes.length; z++){
				System.out.print(QTypes[z] + " ");
			}
			System.out.print("\n  QComponentsBits: ");
			for(int z = 0; z < QComponentsBits.length; z++){
				System.out.print(QComponentsBits[z] + " ");
			}
			for(int z = 0; z < QExponents.length; z++){
				System.out.print("\n  QExponents[" + z + "]: ");
			for(int rLevel = 0; rLevel < QExponents[z].length; rLevel++){
				System.out.print("{");
			for(int subband = 0; subband < QExponents[z][rLevel].length; subband++){
				System.out.print(QExponents[z][rLevel][subband] + ",");
			}
				System.out.print("\b} ");
			}}
			for(int z = 0; z < QMantisas.length; z++){
				System.out.print("\n  QMantisas[" + z + "]: ");
			for(int rLevel = 0; rLevel < QMantisas[z].length; rLevel++){
				System.out.print("{");
			for(int subband = 0; subband < QMantisas[z][rLevel].length; subband++){
				System.out.print(QMantisas[z][rLevel][subband] + ",");
			}
				System.out.print("\b} ");
			}}
			System.out.print("\n  QGuardBits: " + QGuardBits);

			System.out.print("\n  BDBlockWidths: ");
			for(int z = 0; z < BDBlockWidths.length; z++){
				System.out.print(BDBlockWidths[z] + " ");
			}
			System.out.print("\n  BDBlockHeights: ");
			for(int z = 0; z < BDBlockHeights.length; z++){
				System.out.print(BDBlockHeights[z] + " ");
			}

			for(int z = 0; z < BDResolutionPrecinctWidths.length; z++){
				System.out.print("\n  BDResolutionPrecinctWidths[" + z + "]: ");
			for(int rLevel = 0; rLevel < BDResolutionPrecinctWidths[z].length; rLevel++){
					System.out.print("{" + BDResolutionPrecinctWidths[z][rLevel] + "} ");
			}}

			for(int z = 0; z < BDResolutionPrecinctHeights.length; z++){
				System.out.print("\n  BDResolutionPrecinctHeights[" + z + "]: ");
			for(int rLevel = 0; rLevel < BDResolutionPrecinctHeights[z].length; rLevel++){
				System.out.print("{" + BDResolutionPrecinctHeights[z][rLevel] + "} ");
			}}

			System.out.print("\n  LCAchievedNumLayers: " + LCAchievedNumLayers);
			System.out.print("\n  File header type: " + CFileHeader);
			System.out.print("\n  FWProgressionOrder: " + CFileProgression);
			System.out.print("\n");
		}
	}

}
