/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode.FileWrite;
import BOIStream.ByteStream;


/**
 * This class generates the JP2 headers from the JPEG2000 parameters. Usage example:<br>
 * &nbsp; construct<br>
 * &nbsp; run<br>
 *
 * @author Joan Bartrina-Rapesta, Francesc Auli-Llinas
 * @version 1.0
 */
public class JP2Heading{

	/**
	 * Definition in {@link BOICode.Coder#zSize}
	 */
	int zSize;

	/**
	 * Definition in {@link BOICode.Coder#ySize}
	 */
	int ySize;

	/**
	 * Definition in {@link BOICode.Coder#xSize}
	 */
	int xSize;

	/**
	 * Definition in {@link BOICode.Transform.LevelShift#LSSignedComponents}
	 */
	boolean[] LSSignedComponents = null;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTType}
	 */
	int CTType;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	int[] QComponentsBits = null;

	/**
	 * Contains headers of the .jp2 file.
	 */
	ByteStream JP2FileHeaders;

	//INTERNAL VARIABLES

	/**
	 * Tells if different components have different bit depths.
	 */
	boolean bppVariation;


	/**
	 * Constructor of JP2Heading. It receives the information about the compressed image needed to be put in JP2Heading.
	 *
	 * @param zSize definition in {@link BOICode.Coder#zSize}
	 * @param ySize definition in {@link BOICode.Coder#ySize}
	 * @param xSize definition in {@link BOICode.Coder#xSize}
	 * @param LSSignedComponents definition in {@link BOICode.Transform.LevelShift} (see source code)
	 * @param CTType definition in {@link BOICode.Transform.ColourTransform#CTType}
	 * @param QComponentsBits definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	public JP2Heading(int zSize, int ySize, int xSize, boolean[] LSSignedComponents, int CTType, int[] QComponentsBits){
		this.zSize = zSize;
		this.ySize = ySize;
		this.xSize = xSize;
		this.LSSignedComponents = LSSignedComponents;
		this.CTType = CTType;
		this.QComponentsBits = QComponentsBits;
		this.JP2FileHeaders = new ByteStream();
	}

	/**
	 * Generates the JP2 file headers.
	 *
	 * @return the generate ByteStream containing the header
	 */
	public ByteStream run(){
		byte[] marker = null;
		bppVariation = false;

		for(int z=1; z<QComponentsBits.length; z++){
			bppVariation = (QComponentsBits[z] != QComponentsBits[0]);
		}

		//Adds JPEG2000 Signature Box
		marker = genSignatureBox();
		for (int b=0; b<marker.length; b++){
			JP2FileHeaders.addByte(marker[b]);
		}

		//Adds File Type Box
		marker = genFileTypeBox();
		for (int b=0; b<marker.length; b++){
			JP2FileHeaders.addByte(marker[b]);
		}

		//Generates Image Header Box
		byte[] IHB = genImageHeaderBox(bppVariation);

		//Generates Bits Per Component Box
		byte[] BPCB = genBitsPerComponentBox(bppVariation);

		//Generates Colour Specification Box
		byte[] CSB = genColourSpecificationBox();

		//Adds JP2 Header SuperBox
		marker = genJP2HeaderSuperBox(IHB,BPCB,CSB,bppVariation);
		for (int b=0; b<marker.length; b++){
			JP2FileHeaders.addByte(marker[b]);
		}

		//Adds Contiguous Codestream Box (only length and type; content is JPC)
		marker = genContiguousCodestreamBox();
		for (int b=0; b<marker.length; b++){
			JP2FileHeaders.addByte(marker[b]);
		}

		return(JP2FileHeaders);
	}

	/**
	 * Generates the Signature Box.
	 *
	 * @return signature box
	 */
	byte[] genSignatureBox(){
		int SBLength = 12;
		byte[] SignatureBox = new byte[SBLength];
		  //length
		SignatureBox = int2byteArray(SBLength,SignatureBox,0,4);
		  //type = 'jP\040\040'
		SignatureBox[4] = (byte) 0x6A;
		SignatureBox[5] = (byte) 0x50;
		SignatureBox[6] = (byte) 0x20;
		SignatureBox[7] = (byte) 0x20;
		  //content = '<CR><LF><0x87><LF>
		SignatureBox[8] = (byte) 0x0D;
		SignatureBox[9] = (byte) 0x0A;
		SignatureBox[10] = (byte) 0x87;
		SignatureBox[11] = (byte) 0x0A;

		return(SignatureBox);
	}

	/**
	 * Generates the File Type Box.
	 *
	 * @return file type box
	 */
	byte[] genFileTypeBox(){
		int FTBLength = 20;
		byte[] FileTypeBox = new byte[FTBLength];
		  //length
		FileTypeBox = int2byteArray(FTBLength,FileTypeBox,0,4);
		  //type = 'ftyp'
		FileTypeBox[4] = (byte) 0x66;
		FileTypeBox[5] = (byte) 0x74;
		FileTypeBox[6] = (byte) 0x79;
		FileTypeBox[7] = (byte) 0x70;
		  //brand = 'jp2\040'
		FileTypeBox[8] = (byte) 0x6A;
		FileTypeBox[9] = (byte) 0x70;
		FileTypeBox[10] = (byte) 0x32;
		FileTypeBox[11] = (byte) 0x20;
		  //minor version = 0
		FileTypeBox[12] = (byte) 0x00;
		FileTypeBox[13] = (byte) 0x00;
		FileTypeBox[14] = (byte) 0x00;
		FileTypeBox[15] = (byte) 0x00;
		  //CLzero = 'jp2\040' -> ISO 15444-1 (JPEG2000)
		FileTypeBox[16] = (byte) 0x6A;
		FileTypeBox[17] = (byte) 0x70;
		FileTypeBox[18] = (byte) 0x32;
		FileTypeBox[19] = (byte) 0x20;

		return(FileTypeBox);
	}

	/**
	 * Generates the Image Header Box.
	 *
	 * @param bppVariation definition in {@link #bppVariation}
	 * @return image header box
	 */
	byte[] genImageHeaderBox(boolean bppVariation){
		int IHBLength = 22;
		byte[] ImageHeaderBox = new byte[IHBLength];
		//length
		ImageHeaderBox = int2byteArray(IHBLength,ImageHeaderBox,0,4);
		//type
		ImageHeaderBox[4] = (byte) 0x69;
		ImageHeaderBox[5] = (byte) 0x68;
		ImageHeaderBox[6] = (byte) 0x64;
		ImageHeaderBox[7] = (byte) 0x72;
		//image heigth
		ImageHeaderBox = int2byteArray(ySize,ImageHeaderBox,8,4);
		//image width
		ImageHeaderBox = int2byteArray(xSize,ImageHeaderBox,12,4);
		//number of components
		ImageHeaderBox = int2byteArray(zSize,ImageHeaderBox,16,2);
		//bits per component
		if(bppVariation){
			ImageHeaderBox[18] = (byte) 0xFF;
		}
		else{
			int QComponentsByte = QComponentsBits[0]-1;
			if(LSSignedComponents[0]){
				QComponentsByte = changeSignByte(QComponentsByte);
			}
			ImageHeaderBox[18] = (byte) QComponentsByte; //Ssiz
		}
		//compression type
		ImageHeaderBox[19] = (byte) 0x07;
		//unknown colourspace (= false)
		ImageHeaderBox[20] = (byte) 0x00;
		//intellectual property (= no)
		ImageHeaderBox[21] = (byte) 0x00;

		return(ImageHeaderBox);
	}

	/**
	 * Generates the Bits Per Component Box.
	 *
	 * @param bppVariation definition in {@link #bppVariation}
	 * @return bits per component box
	 */
	byte[] genBitsPerComponentBox(boolean bppVariation){
		int BPCBLength = 0;
		byte[] BitsPerComponentBox = null;
		if(bppVariation){
			BPCBLength = 8 + zSize;
			BitsPerComponentBox = new byte[BPCBLength];
			//length
			BitsPerComponentBox = int2byteArray(BPCBLength,BitsPerComponentBox,0,4);
			//type = 'bpcc'
			BitsPerComponentBox[4] = (byte) 0x62;
			BitsPerComponentBox[5] = (byte) 0x70;
			BitsPerComponentBox[6] = (byte) 0x63;
			BitsPerComponentBox[7] = (byte) 0x63;
			//Ssiz for each component
			for (int z=0; z<zSize; z++){
				int QComponentsByte = QComponentsBits[z]-1;
				if(LSSignedComponents[z]){
					QComponentsByte = changeSignByte(QComponentsByte);
				}
				BitsPerComponentBox[8+z] = (byte) QComponentsByte; //Ssiz
			}
		}

		return(BitsPerComponentBox);
	}

	/**
	 * Generates the Colour Specification Box.
	 *
	 * @return colour specification box
	 */
	byte[] genColourSpecificationBox(){
		int CSBLength = 15;
		byte[] ColourSpecificationBox = new byte[CSBLength];
		//length
		ColourSpecificationBox = int2byteArray(CSBLength,ColourSpecificationBox,0,4);
		//type = 'colr'
		ColourSpecificationBox[4] = (byte) 0x63;
		ColourSpecificationBox[5] = (byte) 0x6F;
		ColourSpecificationBox[6] = (byte) 0x6C;
		ColourSpecificationBox[7] = (byte) 0x72;
		//METH (= 1, enumerated colourspace)
		ColourSpecificationBox[8] = (byte) 0x01;
		//precedence (= 0)
		ColourSpecificationBox[9] = (byte) 0x00;
		//colourspace approximation (= 0)
		ColourSpecificationBox[10] = (byte) 0x00;
		//EnumCS
		if(CTType!=0){
			ColourSpecificationBox = int2byteArray(16,ColourSpecificationBox,11,4);//Colour transform applied -> sRGB
		}
		else{
			ColourSpecificationBox = int2byteArray(17,ColourSpecificationBox,11,4);//Colour transform not applied -> greyscale
		}

		return(ColourSpecificationBox);
	}

	/**
	 * Generates the JP2 Header SuperBox.
	 *
	 * @param ImageHeaderBox image header box
	 * @param BitsPerComponentBox bits per component box
	 * @param ColourSpecificationBox colour specification box
	 * @param bppVariation definition in {@link #bppVariation}
	 * @return JP2 header superbox
	 */
	byte[] genJP2HeaderSuperBox(byte[] ImageHeaderBox, byte[] BitsPerComponentBox, byte[] ColourSpecificationBox, boolean bppVariation){
		int JP2HSBLength = 8 + ImageHeaderBox.length + (bppVariation ? BitsPerComponentBox.length:0) + ColourSpecificationBox.length;
		byte[] JP2HeaderSuperBox = new byte[JP2HSBLength];
		//length
		JP2HeaderSuperBox = int2byteArray(JP2HSBLength,JP2HeaderSuperBox,0,4);
		//type = 'jp2h'
		JP2HeaderSuperBox[4] = (byte) 0x6A;
		JP2HeaderSuperBox[5] = (byte) 0x70;
		JP2HeaderSuperBox[6] = (byte) 0x32;
		JP2HeaderSuperBox[7] = (byte) 0x68;
		//content = ImageHeaderBox + (BitsPerComponentBox) + ColourSpecificationBox
		int offset = 8;
		for (int b=0; b<ImageHeaderBox.length; b++){
			JP2HeaderSuperBox[offset+b] = ImageHeaderBox[b];
		}
		offset += ImageHeaderBox.length;
		if(bppVariation){
			for (int b=0; b<BitsPerComponentBox.length; b++){
				JP2HeaderSuperBox[offset+b] = BitsPerComponentBox[b];
			}
		}
		offset += (bppVariation ? BitsPerComponentBox.length: 0);
		for (int b=0; b<ColourSpecificationBox.length; b++){
			JP2HeaderSuperBox[offset+b] = ColourSpecificationBox[b];
		}

		return(JP2HeaderSuperBox);
	}

	/**
	 * Generates the Contiguous Codestream Box (except content, which is the JPEG2000 codestream).
	 *
	 * @return contiguous codestream box
	 */
	byte[] genContiguousCodestreamBox(){
		byte[] ContiguousCodestreamBox = new byte[8];
		//length
		int CCBLength = 0;//not known at the moment of writing these headers -> this box will contain all bytes up to the end of the file
		ContiguousCodestreamBox = int2byteArray(CCBLength,ContiguousCodestreamBox,0,4);
		//type = 'jp2c'
		ContiguousCodestreamBox[4] = (byte) 0x6A;
		ContiguousCodestreamBox[5] = (byte) 0x70;
		ContiguousCodestreamBox[6] = (byte) 0x32;
		ContiguousCodestreamBox[7] = (byte) 0x63;

		return(ContiguousCodestreamBox);
	}

	/**
	 * Forces the first bit of a byte to be one (works with integers).
	 *
	 * @param unsigned unsigned integer representation of a byte
	 * @return signed integer representation of the byte once signed
	 */
	public int changeSignByte(int unsigned){
		int mask1 = 0x0000007F;
		int mask2 = 0x00000080;
		int signed = (mask2 | (unsigned & mask1));
		return (signed);
	}

	/**
	 * Converts an integer number to any desired number of bytes between 1 and 4, and puts them into the specified byte array starting from the specified position.
	 *
	 * @param value integer number to convert to bytes
	 * @param byteArray byte array to be filled
	 * @param position position of the byte array from which the array will be filled
	 * @param bytesNum number of bytes to be put into the byte array
	 * @return byte array filled
	 */
	byte[] int2byteArray(int value, byte[] byteArray, int position, int bytesNum){
		int mask = 0x000000FF;
		byte[] bytes = new byte[bytesNum];
		for (int b=0; b<bytesNum; b++){
			bytes[b] = (byte)((value & (mask << (8*(bytesNum-b-1)))) >> (8*(bytesNum-b-1)));
			byteArray[position+b] = bytes[b];
		}
		return(byteArray);
	}

}