/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode.FileWrite;
import BOIException.WarningException;
import BOIStream.BitStream;
import BOIStream.ByteStream;


/**
 * This class generates the JPC headers from the JPEG2000 parameters. Usage example:<br>
 * &nbsp; construct<br>
 * &nbsp; run<br>
 *
 * @author Joan Bartrina-Rapesta, Francesc Auli-Llinas
 * @version 1.1
 */
public class JPCHeading{

	/**
	 * Definition in {@link BOICode.Coder#zSize}
	 */
	int zSize;
	final int zSize_MAX = 16384;
	final int zSize_BITS = 16;

	/**
	 * Definition in {@link BOICode.Coder#ySize}
	 */
	int ySize;
	final int ySize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#xSize}
	 */
	int xSize;
	final int xSize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#YTOsize}
	 */
	int YTOsize;
	final int YTOsize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#XTOsize}
	 */
	int XTOsize;
	final int XTOsize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#YOsize}
	 */
	int YOsize;
	final int YOsize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#XOsize}
	 */
	int XOsize;
	final int XOsize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#TileYSize}
	 */
	int TileYSize;
	final int TileYSize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#TileXSize}
	 */
	int TileXSize;
	final int TileXSize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Transform.LevelShift#LSSignedComponents}
	 */
	boolean[] LSSignedComponents = null;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTType}
	 */
	int CTType;
	final int CTType_BITS = 8;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 */
	int[] WTTypes = null;
	final int WTTypes_BITS = 8;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 */
	int[] WTLevels = null;
	final int WTLevels_MAX = 32;
	final int WTLevels_BITS = 8;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QTypes}
	 */
	int[] QTypes = null;
	final int QTypes_BITS = 5;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	int[] QComponentsBits = null;
	final int QComponentsBits_BITS = 7;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QExponents}
	 */
	int[][][] QExponents = null;
	final int QExponents_BITS = 5;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QMantisas}
	 */
	int[][][] QMantisas = null;
	final int QMantisas_BITS = 11;

	/**
	 * Definition in {@link BOICode.Transform.Quantization} (see source code)
	 */
	int QGuardBits;
	final int QGuardBits_BITS = 3;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 */
	int[] BDBlockHeights = null;
	final int BDBlockHeights_MAX = 10;
	final int BDBlockHeights_MIN = 2;
	final int BDBlockHeights_BITS = 8;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 */
	int[] BDBlockWidths = null;
	final int BDBlockWidths_MAX = 10;
	final int BDBlockWidths_MIN = 2;
	final int BDBlockWidths_BITS = 8;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctHeights}
	 */
	int[][] BDResolutionPrecinctHeights = null;
	final int BDResolutionPrecinctHeights_BITS = 4;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctWidths}
	 */
	int[][] BDResolutionPrecinctWidths = null;
	final int BDResolutionPrecinctWidths_BITS = 4;

	/**
	 * Definition in {@link BOICode.Code.MQCoder} (see source code)
	 */
	boolean[] MQCFlags = null;

	/**
	 * Definition in {@link BOICode.RateDistortion.LayerCalculation#LCAchievedNumLayers}
	 */
	int LCAchievedNumLayers;
	final int LCAchievedNumLayers_BITS = 16;

	/**
	 * Definition in {@link BOICode.FileWrite.FileWrite#FWProgressionOrder}
	 */
	int FWProgressionOrder;
	final int FWProgressionOrder_BITS = 8;

	/**
	 * Definition in {@link BOICode.FileWrite.FileWrite#FWPacketHeaders}
	 */
	boolean[] FWPacketHeaders = null;

	/**
	 * Contains headers of the JPC file.
	 */
	ByteStream JPCFileHeaders = null;

	//INTERNAL VARIABLES
	/**
	 * Tells if precincts are defined within the headers or not.
	 *
	 * Index in the array is component index
	 */
	boolean[] definedPrecincts = null;

	/**
	 * Denotes capabilities that a decoder needs to properly decode the codestream
	 *
	 */
	int Rsiz;
	final int Rsiz_BITS = 16;

	/**
	 * Horitzontal separation of sample of ith component with respect to the reference grid. thre is one occurence of this paramter for each compoenent
	 *
	 */
	int XRsiz;
	final int XRsiz_BITS = 8;

	/**
	 * Horitzontal separation of sample of ith component with respect to the reference grid. thre is one occurence of this paramter for each compoenent
	 *
	 */
	int YRsiz;
	final int YRsiz_BITS = 8;

	/**
	 * Length of Image and Tile size marker (SIZ) segments in bytes
	 *
	 */
	int LSIZ;
	final int LSIZ_MAX = 49190;
	final int LSIZ_BITS = 16;

	/**
	 * Length of Image and Tile size marker (COD) segments in bytes
	 *
	 */
	int LCOD;
	final int LCOD_MAX = 45;
	final int LCOD_BITS = 16;

	/**
	 * Length of Component Image and Tile size marker (COC) segments in bytes
	 *
	 */
	int LCOC;
	int LCOC_MAX = 43;
	final int LCOC_BITS = 16;

	/**
	 * The length of index of the component to which COC marker segment relates
	 *
	 */
	final int CCOC_MAX = 16383;
	final int CCOC_BITS = 16;

	/**
	 * The length of index of the component to which QCD marker segment relates
	 *
	 */
	int LQCD;
	final int LQCD_MAX = 197;
	final int LQCD_BITS = 16;

	/**
	 * The length of index of the component to which QCC marker segment relates
	 *
	 */
	int LQCC;
	final int LQCC_MAX = 199;
	final int LQCC_BITS = 16;

	/**
	 * The length of index of the component to which QCC marker segment relates
	 *
	 */
	final int CQCC_MAX = 16383;
	final int CQCC_BITS = 16;


	/**
	 * Number of tiles in the image, from 0 to 65534.
	 */
	int tiles;

	/**
	 * Number of tile-parts in each tile. Index of array means tile index (0...65534).
	 *<p>
	 * Values must be between 0 - 254.
	 */
	int[] tileParts = null;


	/**
	 * Constructor of JPCHeading. It receives the information about the compressed image needed to be put in JPCHeading.
	 *
	 * @param zSize definition in {@link BOICode.Coder#zSize}
	 * @param ySize definition in {@link BOICode.Coder#ySize}
	 * @param xSize definition in {@link BOICode.Coder#xSize}
	 * @param TileYSize definition in {@link BOICode.Coder#TileYSize}
	 * @param TileXSize definition in {@link BOICode.Coder#TileXSize}
	 * @param LSSignedComponents definition in {@link BOICode.Transform.LevelShift} (see source code)
	 * @param CTType definition in {@link BOICode.Transform.ColourTransform#CTType}
	 * @param WTTypes definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 * @param WTLevels definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 * @param QTypes definition in {@link BOICode.Transform.Quantization#QTypes}
	 * @param QComponentsBits definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 * @param QGuardBits definition in {@link BOICode.Transform.Quantization} (see source code)
	 * @param QExponents definition in {@link BOICode.Transform.Quantization#QExponents}
	 * @param QMantisas definition in {@link BOICode.Transform.Quantization#QMantisas}
	 * @param BDBlockHeights definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 * @param BDBlockWidths definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 * @param BDResolutionPrecinctHeights definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctHeights}
	 * @param BDResolutionPrecinctWidths definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctWidths}
	 * @param MQCFlags definition in {@link BOICode.Code.MQCoder} (see source code)
	 * @param LCAchievedNumLayers definition in {@link BOICode.RateDistortion.LayerCalculation#LCAchievedNumLayers}
	 * @param FWProgressionOrder definition in {@link BOICode.FileWrite.FileWrite#FWProgressionOrder}
	 * @param FWPacketHeaders definition in {@link BOICode.FileWrite.FileWrite#FWPacketHeaders}
	 */
	public JPCHeading(
	int zSize,
	int ySize,
	int xSize,
	int TileYSize,
	int TileXSize,
	boolean[] LSSignedComponents,
	int CTType,
	int[] WTTypes,
	int[] WTLevels,
	int[] QTypes,
	int[] QComponentsBits,
	int QGuardBits,
	int[][][] QExponents,
	int[][][] QMantisas,
	int[] BDBlockHeights,
	int[] BDBlockWidths,
	int[][] BDResolutionPrecinctHeights,
	int[][] BDResolutionPrecinctWidths,
	boolean[] MQCFlags,
	int LCAchievedNumLayers,
	int FWProgressionOrder,
	boolean[] FWPacketHeaders
	){
		//Parameters copy
		this.zSize = zSize;
		this.ySize = ySize;
		this.xSize = xSize;
		YTOsize = 0;
		XTOsize = 0;
		YOsize = 0;
		XOsize = 0;
		this.TileYSize = TileYSize;
		this.TileXSize = TileXSize;
		this.LSSignedComponents = LSSignedComponents;
		this.CTType = CTType;
		this.WTTypes = WTTypes;
		this.WTLevels = WTLevels;
		this.QTypes = QTypes;
		this.QComponentsBits = QComponentsBits;
		this.QGuardBits = QGuardBits;
		this.QExponents = QExponents;
		this.QMantisas = QMantisas;
		this.BDBlockHeights = BDBlockHeights;
		this.BDBlockWidths = BDBlockWidths;
		this.BDResolutionPrecinctHeights = BDResolutionPrecinctHeights;
		this.BDResolutionPrecinctWidths = BDResolutionPrecinctWidths;
		this.MQCFlags = MQCFlags;
		this.LCAchievedNumLayers = LCAchievedNumLayers;
		this.FWProgressionOrder = FWProgressionOrder;
		this.FWPacketHeaders = FWPacketHeaders;
		tiles = 1;
		tileParts = new int[1];
		tileParts[0] = 1;
		definedPrecincts = new boolean[zSize];
		for (int z=0; z<zSize; z++){
			definedPrecincts[z] = true;
		}
	}

	/**
	 * Generates the JPC file headers.
	 *
	 * @return a bytestream with the JPC markers
	 *
	 * @throws WarningException when the header cannot be generated
	 */
	public ByteStream run() throws WarningException{
		JPCFileHeaders = new ByteStream();
		ByteStream SOCHeading = null;
		ByteStream SIZHeading = null;
		ByteStream CODHeading = null;
		ByteStream COCHeading = null;
		ByteStream QCDHeading = null;
		ByteStream QCCHeading = null;
		//Adds SOC marker
		SOCHeading = genSOC();
		JPCFileHeaders.addBytes(SOCHeading.getByteStream(), SOCHeading.getNumBytes());
		SOCHeading = null;
		//Adds SIZ marker segment
		SIZHeading = genSIZ();
		JPCFileHeaders.addBytes(SIZHeading.getByteStream(), SIZHeading.getNumBytes());
		SIZHeading = null;
		//Adds COD marker segment
		CODHeading = genCOD();
		JPCFileHeaders.addBytes(CODHeading.getByteStream(),CODHeading.getNumBytes());
		CODHeading = null;
		//Adds as many COC marker segments as necessary
		for(int z=1;z<zSize;z++){
			if(neddedCOC(z)){
				COCHeading = genCOC(z);
				JPCFileHeaders.addBytes(COCHeading.getByteStream(),COCHeading.getNumBytes());
			}
		}
		COCHeading = null;
		//Adds QCD marker segment
		QCDHeading = genQCD();
		JPCFileHeaders.addBytes(QCDHeading.getByteStream(),QCDHeading.getNumBytes());
		QCDHeading = null;
		//Adds as many QCC marker segments as necessary
		for(int z=1;z<zSize;z++){
			if(neddedQCC(z)){
				QCCHeading = genQCC(z);
				JPCFileHeaders.addBytes(QCCHeading.getByteStream(),QCCHeading.getNumBytes());

			}
		}
		QCCHeading = null;
		return(JPCFileHeaders);
	}

	/**
	 * Generates the SOC marker (0xFF4F, defined by JPEG2000).
	 *
	 * @return bytestream with the SOC marker
	 */
	ByteStream genSOC(){
		ByteStream SOC = new ByteStream(2);
		SOC.addByte((byte) 0xFF);
		SOC.addByte((byte) 0x4F);
		return (SOC);
	}

	/**
	 * Generates the SIZ marker segment.
	 *
	 * @return bytestream with a SIZ marker segment
	 *
	 * @throws WarningException when the header cannot be generated
	 */
	ByteStream genSIZ() throws WarningException{
		LSIZ = 38;
		Rsiz = 0;
		XRsiz = 1;
		YRsiz = 1;
		BitStream SIZBitStream = new BitStream();
		//SIZ marker
		SIZBitStream.addByte((byte) 0xFF);
		SIZBitStream.addByte((byte) 0x51);
		//SIZ marker segment's Lsiz
		LSIZ += (3 * zSize); //length of marker segment (not including the marker SIZ)
		if((LSIZ < 41) || (LSIZ > LSIZ_MAX)){
			throw new WarningException("LSIZ of marker SIZ is wrong.");
		}
		SIZBitStream.addBits(LSIZ,LSIZ_BITS);
		//SIZ marker segment's Rsiz
		if(Rsiz != 0){
			throw new WarningException("Rsiz of marker SIZ is wrong.");
		}
		SIZBitStream.addBits(Rsiz,Rsiz_BITS);
		//SIZ marker segment's xSize and ySize
		if((xSize < 1) || (xSize > (int)Math.pow(2, xSize_BITS)-1)){
			throw new WarningException("xSize is wrong.");
		}
		SIZBitStream.addBits(xSize,xSize_BITS);
		if((ySize < 1) || (ySize > (int)Math.pow(2, ySize_BITS)-1)){
			throw new WarningException("ySize is wrong.");
		}
		SIZBitStream.addBits(ySize,ySize_BITS);
		//SIZ marker segment's XOsiz,YOsiz
		if((XOsize < 0) || (XOsize > (int)Math.pow(2, XOsize_BITS)-2)){
			throw new WarningException("XOsize is wrong.");
		}
		SIZBitStream.addBits(XOsize,XOsize_BITS);
		if((YOsize < 0) || (YOsize > (int)Math.pow(2, YOsize_BITS)-2)){
			throw new WarningException("YOsize is wrong.");
		}
		SIZBitStream.addBits(YOsize,YOsize_BITS);
		//SIZ marker segment's XTsiz, YTsiz
		if((TileXSize < 1) || (TileXSize > (int)Math.pow(2, TileXSize_BITS)-1)){
			throw new WarningException("TileXSize is wrong.");
		}
		SIZBitStream.addBits(TileXSize,TileXSize_BITS);
		if((TileYSize < 1) || (TileYSize > (int)Math.pow(2, TileYSize_BITS)-1)){
			throw new WarningException("TileYSize is wrong.");
		}
		SIZBitStream.addBits(TileYSize,TileYSize_BITS);
		//SIZ marker segment's XTOsiz,YTOsiz
		if((XTOsize < 0) || (XTOsize > (int)Math.pow(2, XTOsize_BITS)-2)){
			throw new WarningException("XTOsize is wrong.");
		}
		SIZBitStream.addBits(XTOsize,XTOsize_BITS);
		if((YTOsize < 0) || (YTOsize > (int)Math.pow(2, YTOsize_BITS)-2)){
			throw new WarningException("YTOsize is wrong.");
		}
		SIZBitStream.addBits(YTOsize,YTOsize_BITS);
		//SIZ marker segment's Csiz
		if((zSize < 1) || (zSize > zSize_MAX)){
			throw new WarningException("zSize is wrong.");
		}
		SIZBitStream.addBits(zSize,zSize_BITS);
		//SIZ marker segment's Sziz, XRsiz, YRsiz for each component
		if( (XRsiz < 1) || (XRsiz > (int)Math.pow(2, XRsiz_BITS)-1)){
			throw new WarningException("XRsiz is wrong.");
		}
		if( (YRsiz < 1) || (YRsiz > (int)Math.pow(2, YRsiz_BITS)-1)){
			throw new WarningException("YRsiz is wrong.");
		}
		for(int z = 0; z < zSize; z++){
			SIZBitStream.addBit(LSSignedComponents[z]);
			if( (QComponentsBits[z]-1 < 1) || (QComponentsBits[z]-1 > 38)){
				throw new WarningException("QComponentsBits is wrong.");
			}
			SIZBitStream.addBits(QComponentsBits[z]-1,QComponentsBits_BITS);
			SIZBitStream.addBits(XRsiz,XRsiz_BITS);
			SIZBitStream.addBits(YRsiz,YRsiz_BITS);
		}

		ByteStream SIZByteStream = new ByteStream(SIZBitStream.getBitStream(), (int) SIZBitStream.getNumBytes());
		return(SIZByteStream);
	}

	/**
	 * Generates the COD marker segment COD has the default coding style for all components; first component is considered to have the default values.
	 *
	 * @return bytestream with a COD marker segment
	 *
	 * @throws WarningException when COD marker segment cannot be generated due to BOI options
	 */
	ByteStream genCOD() throws WarningException{
		LCOD = 12;
		BitStream CODBitStream = new BitStream();
		//COD marker
		CODBitStream.addByte((byte)0xFF);
		CODBitStream.addByte((byte)0x52);
		//COD marker segment's Lcod
		if(definedPrecincts[0]){
			LCOD += WTLevels[0]+1;
		}
		CODBitStream.addBits(LCOD,LCOD_BITS);
		if((LCOD < 12) || (LCOD > LCOD_MAX)){
			throw new WarningException("LCOD of marker COD is wrong.");
		}
		//COD marker segment's Scod
		CODBitStream.addBits(0,5);
		//End of packet header marker
		CODBitStream.addBit(FWPacketHeaders[1]);
		//Start of packet header marker
		CODBitStream.addBit(FWPacketHeaders[0]);
		//Precincts defined below
		CODBitStream.addBit(definedPrecincts[0]);
		//COD marker segment's SGcod
		if((FWProgressionOrder < 0) || (FWProgressionOrder > 4)){
			throw new WarningException("FWProgressionOrder is wrong.");
		}
		CODBitStream.addBits(FWProgressionOrder,FWProgressionOrder_BITS);
		if((LCAchievedNumLayers < 1) || (LCAchievedNumLayers > (int)Math.pow(2, LCAchievedNumLayers_BITS)-1)){
			throw new WarningException("LCAchievedNumLayers is wrong.");
		}
		CODBitStream.addBits(LCAchievedNumLayers,LCAchievedNumLayers_BITS);
		if((CTType < 0) || (CTType > 2)){
			throw new WarningException("CTType is wrong.");
		}
		if(CTType!=0){
			CODBitStream.addBits(1,CTType_BITS);
		}else{
			CODBitStream.addBits(0,CTType_BITS);
		}
		//COD marker segment's SPcod
		if((WTLevels[0] < 0) || (WTLevels[0] > WTLevels_MAX)){
			throw new WarningException("WTLevels is wrong.");
		}
		CODBitStream.addBits(WTLevels[0],WTLevels_BITS);
		//code-block width and height
		if((BDBlockWidths[0] < BDBlockWidths_MIN) || (BDBlockWidths[0] > BDBlockWidths_MAX) || (BDBlockHeights[0] < BDBlockHeights_MIN) || (BDBlockHeights[0] > BDBlockHeights_MAX) || ((BDBlockWidths[0]+BDBlockHeights[0]) > 12)){
			throw new WarningException("Wrong block sizes.");
		}
		CODBitStream.addBits(BDBlockWidths[0]-2, BDBlockWidths_BITS);
		CODBitStream.addBits(BDBlockHeights[0]-2, BDBlockHeights_BITS);
		//style of code-block coding passes
		CODBitStream.addBits(0,2);
		for(int MQCFlag = MQCFlags.length - 1; MQCFlag >= 0; MQCFlag--){
			CODBitStream.addBit(MQCFlags[MQCFlag]);
		}
		//WT type used
		if( (WTTypes[0] != 1) && (WTTypes[0] != 2)){
			throw new WarningException("WTTypes is wrong.");
		}
		if(WTTypes[0] == 1){
			CODBitStream.addBits(1,WTTypes_BITS);
		}else{
			CODBitStream.addBits(0,WTTypes_BITS);
		}

		CODBitStream = COD_COCPrecinctSizes(CODBitStream,0);
		ByteStream CODByteStream = new ByteStream(CODBitStream.getBitStream(), (int) CODBitStream.getNumBytes());
		return(CODByteStream);
	}

	/**
	 * Generates the COC marker segment, COC has the coding style for a specific component.
	 *
	 * @param z component index
	 * @return bytestream with a COC marker segment
	 *
	 * @throws WarningException when COC marker segment cannot be generated due to BOI options
	 */
	ByteStream genCOC(int z) throws WarningException{
		LCOC = 9;
		BitStream COCBitStream = new BitStream();
		COCBitStream.addByte((byte)0xFF);
		COCBitStream.addByte((byte)0x53);
		//COD marker segment's Lcod
		if(definedPrecincts[z]){
			if(zSize < 257){
				LCOC = 10 + WTLevels[z];
			}else{
				LCOC = 11 + WTLevels[z];
			}
		}else{
			 if(zSize >= 257){
				 LCOC++;
			 }
		}
		if((LCOC < 9) || (LCOC > LCOC_MAX)){
			throw new WarningException("LCOC of marker COC is wrong.");
		}
		if(z > CCOC_MAX){
			throw new WarningException("Number of COC's has been exceded.");
		}
		COCBitStream.addBits(LCOC,LCOC_BITS);
		if(zSize < 257){
			COCBitStream.addBits(z,CCOC_BITS/2);
		}else{
			COCBitStream.addBits(z,CCOC_BITS);
		}
		//COC marker segment's Scoc
		COCBitStream.addBits(0,7);
		COCBitStream.addBit(definedPrecincts[z]);
		//COD marker segment's SPcoc
		if((WTLevels[z] < 0) || (WTLevels[z] > WTLevels_MAX)){
			throw new WarningException("WTLevels is wrong.");
		}
		COCBitStream.addBits(WTLevels[z],WTLevels_BITS);
		//code-block width and height
		if((BDBlockWidths[z] < BDBlockWidths_MIN) || (BDBlockWidths[z] > BDBlockWidths_MAX) || (BDBlockHeights[z] < BDBlockHeights_MIN) || (BDBlockHeights[z] > BDBlockHeights_MAX) || ((BDBlockWidths[z]+BDBlockHeights[z]) > 12)){
			throw new WarningException("Wrong block sizes.");
		}
		COCBitStream.addBits(BDBlockWidths[z]-2, BDBlockWidths_BITS);
		COCBitStream.addBits(BDBlockHeights[z]-2, BDBlockHeights_BITS);
		//style of code-block coding passes
		COCBitStream.addBits(0,2);
		for(int MQCFlag = MQCFlags.length - 1; MQCFlag >= 0; MQCFlag--){
			COCBitStream.addBit(MQCFlags[MQCFlag]);
		}
		//WT type used
		if( (WTTypes[z] != 1) && (WTTypes[z] != 2)){
			throw new WarningException("WTTypes is wrong.");
		}
		if(WTTypes[z] == 1){
			COCBitStream.addBits(1,WTTypes_BITS);
		}else{
			COCBitStream.addBits(0,WTTypes_BITS);
		}

		COCBitStream = COD_COCPrecinctSizes(COCBitStream,z);
		ByteStream COCByteStream = new ByteStream(COCBitStream.getBitStream(), (int) COCBitStream.getNumBytes());
		return(COCByteStream);
	}

	/**
	 * Recives the COD or COC BitStream marker and fill it with the precinct Sizes
	 *
	 * @param COD_COCBitstream BitStream to fill
	 * @param z component index
	 * @return bitstream with a COD or COC marker segment
	 *
	 * @throws WarningException when COD or COC marker segment cannot be generated due to BOI options
	 */
	BitStream COD_COCPrecinctSizes(BitStream COD_COCBitstream, int z) throws WarningException{
		int widthExp, heightExp;
		if(definedPrecincts[z]){
			for (int i = 0 ; i < (WTLevels[z] + 1); i++){ //first corresponds to LL subband; each successive index corresponds to each resolution level in order
				heightExp = BDResolutionPrecinctHeights[z][i] + (i == 0 ? 0: 1);
				if( (heightExp < 0) && (heightExp > Math.pow(2,BDResolutionPrecinctHeights_BITS))){
					throw new WarningException("BDResolutionPrecinctHeights is wrong.");
				}
				COD_COCBitstream.addBits(heightExp,BDResolutionPrecinctHeights_BITS);
				widthExp = BDResolutionPrecinctWidths[z][i] + (i == 0 ? 0: 1);
				if( (widthExp < 0) && (widthExp > Math.pow(2,BDResolutionPrecinctWidths_BITS))){
					throw new WarningException("BDResolutionPrecinctWidths is wrong.");
				}
				COD_COCBitstream.addBits(widthExp,BDResolutionPrecinctWidths_BITS);
			}
		}
		return(COD_COCBitstream);
	}


	/**
	 * Generates the QCD marker segment, QCD has the default quantization parameters for all components.
	 *
	 * @return bytestream with a  QCD marker segment
	 *
	 * @throws WarningException when QCD marker segment cannot be generated due to BOI options
	 */
	ByteStream genQCD() throws WarningException{
		LQCD = 4;
		BitStream QCDBitStream = new BitStream();
		//QCD marker
		QCDBitStream.addByte((byte)0xFF);
		QCDBitStream.addByte((byte)0x5C);
		//QCD marker segment's Lqcd
		switch(QTypes[0]){
		case 0:	//no quantization
			LQCD = 4 + 3 * WTLevels[0];
			break;
		case 1:	//scalar quantization derived
			LQCD = 5;
			break;
		case 2:	//scalar quantization expounded
			LQCD = 5 + (6 * WTLevels[0]);
			break;
		}
		if((LQCD < 4) || (LQCD > LQCD_MAX)){
			throw new WarningException("LQCD of marker QCD is wrong.");
		}
		QCDBitStream.addBits(LQCD,LQCD_BITS);
		//QCD marker segment's Sqcd
		//guard bits
		if( (QGuardBits < 0) || (QGuardBits > Math.pow(2,QGuardBits_BITS)-1) ){
			throw new WarningException("QGuardBits of marker QCD is wrong.");
		}
		QCDBitStream.addBits(QGuardBits,QGuardBits_BITS);
		QCDBitStream.addBits(QTypes[0],QTypes_BITS);
		//QCD marker segments' SPqcd
		QCDBitStream = QCD_QCCQuantization(QCDBitStream, 0);

		ByteStream QCDByteStream = new ByteStream(QCDBitStream.getBitStream(), (int) QCDBitStream.getNumBytes());
		return(QCDByteStream);
	}


	/**
	 * Generates the QCC marker segment, QCC has the quantization parameters for a specific component.
	 *
	 * @param z component index
	 *
	 * @return bytestream with a QCC marker segment
	 *
	 * @throws WarningException when QCC marker segment cannot be generated due to BOI options
	 */
	ByteStream genQCC(int z) throws WarningException{
		LQCC = 5;
		BitStream QCCBitStream = new BitStream();
		//QCC marker
		QCCBitStream.addByte((byte)0xFF);
		QCCBitStream.addByte((byte)0x5D);
		//QCD marker segment's Lqcd
		switch(QTypes[0]){
		case 0:	//no quantization
			if(zSize < 257){
				LQCC = 5 + 3 * WTLevels[z];
			}else{
				LQCC = 6 + 3 * WTLevels[z];
			}
			break;
		case 1:	//scalar quantization derived
			if(zSize < 257){
				LQCC = 6;
			}else{
				LQCC = 7;
			}

			break;
		case 2:	//scalar quantization expounded
			if(zSize < 257){
				LQCC = 6 + 6 * WTLevels[z];
			}else{
				LQCC = 7 + 6 * WTLevels[z];
			}
			break;
		default:
			throw new WarningException("Unrecognized quantization type.");
		}
		if(LQCC > LQCC_MAX){
			throw new WarningException("Length of QCC is wrong.");
		}
		QCCBitStream.addBits(LQCC,LQCC_BITS);
		//Cqcc component index
		//zSize_MAX
		if(z > CQCC_MAX){
			throw new WarningException("Numbers of QCC markers has been exceded.");
		}
		if(zSize < 257){
			QCCBitStream.addBits(z,CQCC_BITS/2);
		}else{
			QCCBitStream.addBits(z,CQCC_BITS);
		}
		//QCD marker segment's Sqcc
		if( (QGuardBits < 0) || (QGuardBits > Math.pow(2,QGuardBits_BITS)-1) ){
			throw new WarningException("QGuardBits of marker QCD is wrong.");
		}
		QCCBitStream.addBits(QGuardBits,QGuardBits_BITS);
		QCCBitStream.addBits(QTypes[z],QTypes_BITS);
		//QCC marker segments' SPqcc
		QCCBitStream = QCD_QCCQuantization(QCCBitStream, z);

		ByteStream QCCByteStream = new ByteStream(QCCBitStream.getBitStream(), (int) QCCBitStream.getNumBytes());
		return(QCCByteStream);
	}

	/**
	 * Recives the COD or COC BitStream marker and fill it with the precinct Sizes
	 *
	 * @param QCD_QCCQuantization BitStream to fill
	 * @param z component index
	 *
	 * @return bitstream with a COD or COC marker segment
	 *
	 * @throws WarningException when COD or COC marker segment cannot be generated due to BOI options
	 */
	BitStream QCD_QCCQuantization(BitStream QCD_QCCQuantization, int z)throws WarningException{
		switch(QTypes[z]){
		case 0: //no quantization
			if ((QExponents[z][0][0] < 0) || (QExponents[z][0][0] > Math.pow(2,QExponents_BITS)-1)){
				throw new WarningException("Invalid exponent.");
			}
			QCD_QCCQuantization.addBits(QExponents[z][0][0],QExponents_BITS);
			QCD_QCCQuantization.addBits(0,3);
			for(int rLevel = 1; rLevel <= WTLevels[z]; rLevel++){
				for(int subband = 0; subband < QExponents[z][rLevel].length; subband++){
					if ((QExponents[z][rLevel][subband] < 0) || (QExponents[z][rLevel][subband] > Math.pow(2,QExponents_BITS)-1)){
						throw new WarningException("Invalid exponent.");
					}
					QCD_QCCQuantization.addBits(QExponents[z][rLevel][subband],QExponents_BITS);
					QCD_QCCQuantization.addBits(0,3);
				}
			}
			break;
		case 1: //scalar quatization derived
			if ((QExponents[z][0][0] < 0) || (QExponents[z][0][0] > Math.pow(2,QExponents_BITS)-1) || (QMantisas[z][0][0] < 0) || QMantisas[z][0][0] > Math.pow(2,QMantisas_BITS)-1){
				throw new WarningException("Invalid exponent or Mantisa.");
			}
			QCD_QCCQuantization.addBits(QExponents[z][0][0],QExponents_BITS);
			QCD_QCCQuantization.addBits(QMantisas[z][0][0],QMantisas_BITS);
			break;
		case 2: //scalar quantization expounded
			for(int rLevel = 0; rLevel <= WTLevels[z]; rLevel++){
				for(int subband = 0; subband < QExponents[z][rLevel].length; subband++){
					if ((QExponents[z][rLevel][subband] < 0) || (QExponents[z][rLevel][subband] > Math.pow(2,QExponents_BITS)-1) || (QMantisas[z][rLevel][subband] < 0) || QMantisas[z][rLevel][subband] > Math.pow(2,QMantisas_BITS)-1){
						throw new WarningException("Invalid exponent or Mantisa.");
					}
					QCD_QCCQuantization.addBits(QExponents[z][rLevel][subband],QExponents_BITS);
					QCD_QCCQuantization.addBits(QMantisas[z][rLevel][subband],QMantisas_BITS);
				}
			}
				break;
		}
		return(QCD_QCCQuantization);
	}

	/**
	 * To know if a COC marker segment is needed for a specific component.
	 *
	 * @param component component index
	 *
	 * @return true or false
	 */
	boolean neddedCOC(int component){
		boolean COCneeded = false;
		if((WTLevels[component] != WTLevels[0]) || (WTTypes[component] != WTTypes[0]) || (BDBlockWidths[component] != BDBlockWidths[0]) || (BDBlockHeights[component] != BDBlockHeights[0] )){
			COCneeded = true;
		}else{
			for(int rLevel = 0; rLevel < (WTLevels[component] + 1); rLevel++){
				if((BDResolutionPrecinctHeights[component][rLevel] != BDResolutionPrecinctHeights[0][rLevel]) || (BDResolutionPrecinctWidths[component][rLevel] != BDResolutionPrecinctWidths[0][rLevel] )){
					COCneeded = true;
					rLevel = WTLevels[component]+1;
				}
			}
		}
		return(COCneeded);
	}

	/**
	 * To know if a QCC marker segment is needed for a specific component.
	 *
	 * @param component component index
	 *
	 * @return true or false
	 */
	boolean neddedQCC(int component){
		boolean QCCneeded = false;
		if((WTLevels[component] != WTLevels[0])||(QTypes[component] != QTypes[0])){
			QCCneeded = true;
		}else{
			for(int rLevel = 0; rLevel < QExponents[component].length; rLevel++){
				for(int subband = 0; subband < QExponents[component][rLevel].length; subband++){
					if(QExponents[component][rLevel][subband] != QExponents[0][rLevel][subband]){
						QCCneeded = true;
					}
				}
			}
		}
		return(QCCneeded);
	}
}
