/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOIDecode.Decode;
import BOIException.ErrorException;
import BOIException.WarningException;
import BOIStream.ByteStream;
import BOICode.Code.CoderLUTS;


/**
 * This class receives an array of ByteStreams belonging to a code-block and decodes them obtaining the original image samples.
 *
 * @author Francesc Auli-Llinas
 * @version 1.4
 */
public class BPEDecoder{
	final static int maxSize = 128;
	final static int yCell = 8;
	final static int xCell = 8;
	int maxCellsInPass = 1;

	int[][] approxSamples = new int[maxSize][maxSize];
	short[][] statusMap = new short[maxSize][maxSize];

	int[][] listREFS = new int[maxSize * maxSize][2];

	int ySizeCellsMap = -1;
	int xSizeCellsMap = -1;
	int sizeListCells = -1;
	int firstNonVisited = -1;
	int[][] listCellsSIGS = new int[(maxSize/yCell) * (maxSize/xCell)][3];
	int[][] cellsMapCoordinates = new int[(maxSize/yCell) * (maxSize/xCell)][4];

	int LSBPlane = -1;
	short currCellsInPass = -1;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#CDByteStream}
	 */
	ByteStream CDByteStream = null;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#blockSamples}
	 */
	float[][] blockSamples = null;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#contextMap}
	 */
	byte[][] contextMap = null;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#ySize}
	 */
	int ySize = -1;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#xSize}
	 */
	int xSize = -1;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#subband}
	 */
	int subband;

	/**
	 * Most Significance Bit Plane of the subband.
	 * <p>
	 * Negative values not allowed.
	 */
	int subbandMSBPlane;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#MSBPlane}
	 */
	int MSBPlane;

	/**
	 * Definition in {@link BOIDecode.Decode.BlockBuild#BCCodingPasses}
	 */
	int codingPasses;

	/**
	 * Whether to use integer processing or not. Integer processing is commonly employed when the Integer Wavelet Transform is used. The main difference with respect to the normal processing is that with integer processing the decoded coefficients are rounded to the nearest integer rather than to leave them as a floating point number (which is caused when BDreconstruction is a value different from 0.5).
	 * <p>
	 * true indicates that integer processing is prefered.
	 */
	int intProcessing;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#currentBitPlane}
	 */
	private int currentBitPlane;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#currentCodingPass}
	 */
	private int currentCodingPassGlobal;
	private int currentCodingPass;

	static final private int[] currentBitPlaneMask = {
		(1 << 0), (1 << 1), (1 << 2), (1 << 3), (1 << 4), (1 << 5), (1 << 6), (1 << 7), (1 << 8), (1 << 9), (1 << 10), (1 << 11), (1 << 12), (1 << 13), (1 << 14), (1 << 15), (1 << 16), (1 << 17), (1 << 18)
	};
	static final private int[] currentBitPlaneMaskNull = {
		~(currentBitPlaneMask[0]-1), ~(currentBitPlaneMask[1]-1), ~(currentBitPlaneMask[2]-1), ~(currentBitPlaneMask[3]-1), ~(currentBitPlaneMask[4]-1), ~(currentBitPlaneMask[5]-1), ~(currentBitPlaneMask[6]-1), ~(currentBitPlaneMask[7]-1), ~(currentBitPlaneMask[8]-1), ~(currentBitPlaneMask[9]-1), ~(currentBitPlaneMask[10]-1), ~(currentBitPlaneMask[11]-1), ~(currentBitPlaneMask[12]-1), ~(currentBitPlaneMask[13]-1), ~(currentBitPlaneMask[14]-1), ~(currentBitPlaneMask[15]-1), ~(currentBitPlaneMask[16]-1), ~(currentBitPlaneMask[17]-1), ~(currentBitPlaneMask[18]-1)
	};

	/**
	 * Reconsuctrion value for the non transmitted bit of a coefficient.
	 * <p>
	 * Value computed using the above LUTs.
	 */
	private float deqReconstruction = 0.0F;

	private int numSIGS;
	private int numSIGS_CURRBP;
	private int numREFS;
	private int indexListREFS;

	/**
	 * Decoder to read from the inputByteStream.
	 * <p>
	 * This Decoder must be compatible with the abstract class Decoder.
	 */
	private Decoder blockDecoder = null;

	/**
	 * Constructor that does not receive anything. Before use the functions of decodeNextCodingPass the function swapCodeBlock must be used.
	 */
	public BPEDecoder(){
		blockDecoder = new MQDecoder();
	}

	/**
	 * Swaps the code-block decoded by this class. Some initializations performed.
	 *
	 * @param CDByteStream definition in {@link #CDByteStream}
	 * @param blockSamples definition in {@link #blockSamples}
	 * @param MSBPlane definition in {@link BOICode.Code.BPECoder#MSBPlane}
	 * @param codingPasses definition in {@link BOIDecode.Decode.BlockBuild#BCCodingPasses}
	 * @param subband definition in {@link #subband}
	 * @param subbandMSBPlane definition in {@link #subbandMSBPlane}
	 * @param intProcessing definition in {@link #intProcessing}
	 * @param probSIG
	 * @param probREF
	 * @throws ErrorException problems with the inputByteStream
	 */
	public void swapCodeBlock(ByteStream CDByteStream, float[][] blockSamples, int MSBPlane, int codingPasses, int subband, int subbandMSBPlane, boolean intProcessing) throws ErrorException{
		int numBitPlanes = MSBPlane + 1;

		this.blockSamples = blockSamples;
		this.MSBPlane = MSBPlane;
		this.codingPasses = codingPasses;
		this.subband = subband;
		this.subbandMSBPlane = subbandMSBPlane;
		this.intProcessing = intProcessing ? 1 : 0;
		this.CDByteStream = CDByteStream;

		boolean sizeChanged = false;
		if(ySize != blockSamples.length || xSize != blockSamples[0].length){
			this.ySize = blockSamples.length;
			this.xSize = blockSamples[0].length;
			this.ySizeCellsMap = (int)Math.ceil((float)ySize / (float)yCell);
			this.xSizeCellsMap = (int)Math.ceil((float)xSize / (float)xCell);
			this.sizeListCells = ySizeCellsMap * xSizeCellsMap;
			this.contextMap = new byte[ySize][xSize];
			sizeChanged = true;
		}
		int bitPlanes = MSBPlane + 1;

		if(ySize*xSize > 4096){
			maxCellsInPass = 2;
		}

		//Set blockDecoder (we do not create another object, we reutilize the already created decoder)
		blockDecoder.swapInputByteStream(CDByteStream);
		blockDecoder.restart();
		blockDecoder.reset();

		//Status map
		for(int x = 0; x < xSize; x++){
		for(int y = 0; y < ySize; y++){
			statusMap[y][x] &= 0xF000; //reset everything except boundaries
			contextMap[y][x] = 0; //reset context
		}}
		if(sizeChanged){
			for(int x = 0; x < xSize; x++){
			for(int y = 0; y < ySize; y++){
				statusMap[y][x] &= 0x0FFF;
			}}
			statusMap[0][0] |= 0xA000;
			statusMap[ySize-1][0] |= 0x6000;
			statusMap[0][xSize-1] |= 0x9000;
			statusMap[ySize-1][xSize-1] |= 0x5000;
			for(int x = 1; x < xSize-1; x++){
				statusMap[0][x] |= 0x8000;
				statusMap[ySize-1][x] |= 0x4000;
			}
			for(int y = 1; y < ySize-1; y++){
				statusMap[y][0] |= 0x2000;
				statusMap[y][xSize-1] |= 0x1000;
			}

			int yBegin = 0;
			int yEnd = yCell;
			for(int yMap = 0; yMap < ySizeCellsMap; yMap++){
				if(yMap == ySizeCellsMap - 1){
					yEnd = ySize;
				}
				int xBegin = 0;
				int xEnd = xCell;
				for(int xMap = 0; xMap < xSizeCellsMap; xMap++){
					if(xMap == xSizeCellsMap - 1){
						xEnd = xSize;
					}
					int cell = yMap * xSizeCellsMap + xMap;
					cellsMapCoordinates[cell][0] = yBegin;
					cellsMapCoordinates[cell][1] = yEnd;
					cellsMapCoordinates[cell][2] = xBegin;
					cellsMapCoordinates[cell][3] = xEnd;
					xBegin = xEnd;
					xEnd += xCell;
				}
				yBegin = yEnd;
				yEnd += yCell;
			}
		}
		for(int cell = 0; cell < sizeListCells; cell++){
			listCellsSIGS[cell][0] = cell;
			listCellsSIGS[cell][1] = -1;
			listCellsSIGS[cell][2] = 0;
		}

		//Initialize current variables
		currentBitPlane = this.MSBPlane;
		currentCodingPassGlobal = 0;
		currentCodingPass = 0;
		numSIGS = numREFS = numSIGS_CURRBP = indexListREFS = 0;
		currCellsInPass = 0;
		firstNonVisited = 0;
		if(this.intProcessing == 0){
			deqReconstruction = (float)currentBitPlaneMask[currentBitPlane] * DecoderLUTS.BDReconstruction[this.intProcessing][subbandMSBPlane][subbandMSBPlane-currentBitPlane];
		}else{
			deqReconstruction = Math.round ( (float)(currentBitPlaneMask[currentBitPlane]-1) * DecoderLUTS.BDReconstruction[this.intProcessing][subbandMSBPlane][subbandMSBPlane-currentBitPlane] );
		}
	}

	/**
	 * Decode the image samples (all bit planes).
	 *
	 * @throws WarningException when reached the last coding pass
	 * @throws ErrorException problems with the inputByteStream
	 */
	public void runAll() throws WarningException, ErrorException{
		//Decode the LSBPlane
		int numBits = (int) (Math.log( (float) MSBPlane) / Math.log(2.0D));
		numBits = numBits <= 0 ? 1: numBits;
		LSBPlane = 0;
		for(int i = numBits; i >= 0; i--){
			boolean bit = blockDecoder.decodeBit(18);
			if(bit){
				LSBPlane += (1 << i);
			}
		}
		//Decode all coding segments
		while((currentBitPlane >= 0) && (currentCodingPassGlobal < codingPasses)){
			decodeNextPass();
		}
		//Set the sign
		for(int x=0;x<xSize;x++){
		for(int y=0;y<ySize;y++){
			if((blockSamples[y][x] > 0) && (statusMap[y][x] & 0x0400) != 0){
				blockSamples[y][x] = -blockSamples[y][x];
			}
		}}
	}

	public void decodeNextPass() throws ErrorException{
		switch(currentCodingPass){
		case 0:
			if(significanceCoding()){
				if(currCellsInPass == maxCellsInPass){
					currentCodingPassGlobal++;
					currCellsInPass = 0;
					numSIGS_CURRBP += numSIGS;
					numSIGS = 0;
				}
			}else{
				if(currCellsInPass > 0){
					currentCodingPassGlobal++;
					currCellsInPass = 0;
					numSIGS_CURRBP += numSIGS;
					numSIGS = 0;
				}
				currentCodingPass = 1;
			}
			break;
		case 1:
			if(currentBitPlane < MSBPlane){
				refinementCoding();
				currentCodingPassGlobal++;
			}
			currentCodingPass = 0;
			numREFS += numSIGS_CURRBP;
			numSIGS_CURRBP = 0;
			firstNonVisited = 0;
			currentBitPlane--;
			if(currentBitPlane >= 0){
				if(intProcessing == 0){
					deqReconstruction = (float)currentBitPlaneMask[currentBitPlane] * DecoderLUTS.BDReconstruction[intProcessing][subbandMSBPlane][subbandMSBPlane-currentBitPlane];
				}else{
					deqReconstruction = Math.round ( (float)(currentBitPlaneMask[currentBitPlane]-1) * DecoderLUTS.BDReconstruction[intProcessing][subbandMSBPlane][subbandMSBPlane-currentBitPlane] );
				}
			}
			break;
		}
	}

	private void foundSignificant(int x, int y) throws ErrorException{
		//Sign encode
		int signContext = 0;
		switch(statusMap[y][x] & 0x00C0){
		case 0x0040:
			switch(statusMap[y][x] & 0x0300){
			case 0x0100:
				signContext = -14;
				break;
			case 0x0000:
				signContext = -13;
				break;
			case 0x0200:
				signContext = -12;
				break;
			}
			break;
		case 0x0000:
			switch(statusMap[y][x] & 0x0300){
			case 0x0100:
				signContext = -11;
				break;
			case 0x0000:
				signContext = 10;
				break;
			case 0x0200:
				signContext = 11;
				break;
			}
			break;
		case 0x0080:
			switch(statusMap[y][x] & 0x0300){
			case 0x0100:
				signContext = 12;
				break;
			case 0x0000:
				signContext = 13;
				break;
			case 0x0200:
				signContext = 14;
				break;
			}
			break;
		}

		boolean sign = blockDecoder.decodeBit(Math.abs(signContext));
		boolean signNeg = false;
		if(signContext >= 0 == sign){
			//blockSamples[y][x] = -blockSamples[y][x];
			statusMap[y][x] |= 0x0400;
			signNeg = true;
		}

		//Update CPP flag, and sign context
		if(y >= 1){
			statusMap[y-1][x] |= 0x0020; //CPP
			switch(statusMap[y-1][x] & 0x0300){ //SIGN
			case 0x0000:
				if(signNeg) statusMap[y-1][x] |= 0x0100;
				else statusMap[y-1][x] |= 0x0200;
				break;
			case 0x0100:
				if(!signNeg) statusMap[y-1][x] &= 0xFCFF;
				break;
			case 0x0200:
				if(signNeg) statusMap[y-1][x] &= 0xFCFF;
				break;
			}
			if(x >= 1){
				statusMap[y-1][x-1] |= 0x0020;
			}
			//if(x+1 < xSize){
			if((statusMap[y][x] & 0x1000) == 0){
				statusMap[y-1][x+1] |= 0x0020;
			}
		}
		//if(y+1 < ySize){
		if((statusMap[y][x] & 0x4000) == 0){
			statusMap[y+1][x] |= 0x0020;
			switch(statusMap[y+1][x] & 0x0300){
			case 0x0000:
				if(signNeg) statusMap[y+1][x] |= 0x0100;
				else statusMap[y+1][x] |= 0x0200;
			break;
			case 0x0100:
				if(!signNeg) statusMap[y+1][x] &= 0xFCFF;
			break;
			case 0x0200:
				if(signNeg) statusMap[y+1][x] &= 0xFCFF;
			break;
			}
			if(x >= 1){
				statusMap[y+1][x-1] |= 0x0020;
			}
			//if(x+1 < xSize){
			if((statusMap[y][x] & 0x1000) == 0){
				statusMap[y+1][x+1] |= 0x0020;
			}
		}
		if(x >= 1){
			statusMap[y][x-1] |= 0x0020;
			switch(statusMap[y][x-1] & 0x00C0){
			case 0x0000:
				if(signNeg) statusMap[y][x-1] |= 0x0040;
				else statusMap[y][x-1] |= 0x0080;
			break;
			case 0x0040:
				if(!signNeg) statusMap[y][x-1] &= 0xFF3F;
			break;
			case 0x0080:
				if(signNeg) statusMap[y][x-1] &= 0xFF3F;
			break;
			}
		}
		//if(x+1 < xSize){
		if((statusMap[y][x] & 0x1000) == 0){
			statusMap[y][x+1] |= 0x0020;
			switch(statusMap[y][x+1] & 0x00C0){
			case 0x0000:
				if(signNeg) statusMap[y][x+1] |= 0x0040;
				else statusMap[y][x+1] |= 0x0080;
			break;
			case 0x0040:
				if(!signNeg) statusMap[y][x+1] &= 0xFF3F;
			break;
			case 0x0080:
				if(signNeg) statusMap[y][x+1] &= 0xFF3F;
			break;
			}
		}

		//update currentBP status map
		statusMap[y][x] |= currentBitPlane;

		//update list of refinements
		listREFS[indexListREFS][0] = x;
		listREFS[indexListREFS++][1] = y;

		//increase numSIGS
		numSIGS++;
	}

	private boolean significanceCoding() throws ErrorException{
		if(firstNonVisited < sizeListCells){
			boolean scanCell = true;
			if((currentBitPlane > LSBPlane) && (listCellsSIGS[firstNonVisited][1] == -1)){
				scanCell = blockDecoder.decodeBit(9);
				if(scanCell){
					listCellsSIGS[firstNonVisited][1] = currentBitPlane;
				}
			}
			if(scanCell){
				currCellsInPass++;
				int correspondingCell = listCellsSIGS[firstNonVisited][0];
				if(subband == 2){
					for(int y = cellsMapCoordinates[correspondingCell][0]; y < cellsMapCoordinates[correspondingCell][1]; y++){
					for(int x = cellsMapCoordinates[correspondingCell][2]; x < cellsMapCoordinates[correspondingCell][3]; x++){
						if((statusMap[y][x] & 0x001F) == 0){
							boolean significant = false;
							significant = blockDecoder.decodeBit(contextMap[y][x]);
							if(significant){
								blockSamples[y][x] = currentBitPlaneMask[currentBitPlane] + deqReconstruction;
								foundSignificant(x, y);
								updateMap(x, y);
								listCellsSIGS[firstNonVisited][2]++;
							}
						}
					}}
				}else{
					for(int x = cellsMapCoordinates[correspondingCell][2]; x < cellsMapCoordinates[correspondingCell][3]; x++){
					for(int y = cellsMapCoordinates[correspondingCell][0]; y < cellsMapCoordinates[correspondingCell][1]; y++){
						if((statusMap[y][x] & 0x001F) == 0){
							boolean significant = false;
							significant = blockDecoder.decodeBit(contextMap[y][x]);
							if(significant){
								blockSamples[y][x] = currentBitPlaneMask[currentBitPlane] + deqReconstruction;
								foundSignificant(x, y);
								updateMap(x, y);
								listCellsSIGS[firstNonVisited][2]++;
							}
						}
					}}
				}
			}
			sortCell(firstNonVisited);
			firstNonVisited++;
			return(true);
		}else{
			return(false);
		}
	}

	private void refinementCoding() throws ErrorException{
		for(int coef = numREFS - 1; coef >= 0; coef--){
			boolean refinement = blockDecoder.decodeBit(contextMap[listREFS[coef][1]][listREFS[coef][0]]);
			if(refinement){
				blockSamples[listREFS[coef][1]][listREFS[coef][0]] = ( (float)(((int)blockSamples[listREFS[coef][1]][listREFS[coef][0]] & currentBitPlaneMaskNull[currentBitPlane]) | currentBitPlaneMask[currentBitPlane]) + deqReconstruction);
			}else{
				blockSamples[listREFS[coef][1]][listREFS[coef][0]] = ( (float)(((int)blockSamples[listREFS[coef][1]][listREFS[coef][0]] & currentBitPlaneMaskNull[currentBitPlane]) & ~currentBitPlaneMask[currentBitPlane]) + deqReconstruction);
			}
		}
	}

	private void sortCell(int cell){
		int newPosition = cell - 1;
		while((newPosition >= 0) && (listCellsSIGS[newPosition][2] < listCellsSIGS[cell][2])){
			newPosition--;
		}
		newPosition++;
		if(newPosition < cell){
			int[] tmp = listCellsSIGS[cell];
			System.arraycopy(listCellsSIGS, newPosition, listCellsSIGS, newPosition + 1, cell - newPosition);
			listCellsSIGS[newPosition] = tmp;
		}
	}

	/**
	 * @return blockSamples definition in {@link #blockSamples}
	 */
	public float[][] getBlockSamples(){
		return(blockSamples);
	}

	/**
	 * Turn neighbors map samples of (x,y) from 0 to 1 if they can be visited in the current scan.
	 *
	 * @param x x position of the sample
	 * @param y y position of the sample
	 * @param CP indicates if the updating is performed by the Cleanup Pass (true, otherwise false)
	*/
	private void updateMap(int x, int y){
		//Current coefficient must be refined
		contextMap[y][x] = 17;

		//Context of the neighbors
		switch(subband){
		case 0: //SUBBAND LL
		case 2: //SUBBAND LH
			if(x-1 >= 0){
				//LEFT
				switch(contextMap[y][x-1]){
				case 0:
					contextMap[y][x-1] = (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y][x-1] = (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y][x-1] = (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y][x-1] = (byte) 8;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(contextMap[y-1][x-1]){
					case 0:
						contextMap[y-1][x-1] = (byte) 1;
						break;
					case 1:
						contextMap[y-1][x-1] = (byte) 2;
						break;
					case 5:
						contextMap[y-1][x-1] = (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(contextMap[y+1][x-1]){
					case 0:
						contextMap[y+1][x-1] = (byte) 1;
						break;
					case 1:
						contextMap[y+1][x-1] = (byte) 2;
						break;
					case 5:
						contextMap[y+1][x-1] = (byte) 6;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(contextMap[y][x+1]){
				case 0:
					contextMap[y][x+1] = (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y][x+1] = (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y][x+1] = (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y][x+1] = (byte) 8;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(contextMap[y-1][x+1]){
					case 0:
						contextMap[y-1][x+1] = (byte) 1;
						break;
					case 1:
						contextMap[y-1][x+1] = (byte) 2;
						break;
					case 5:
						contextMap[y-1][x+1] = (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(contextMap[y+1][x+1]){
					case 0:
						contextMap[y+1][x+1] = (byte) 1;
						break;
					case 1:
						contextMap[y+1][x+1] = (byte) 2;
						break;
					case 5:
						contextMap[y+1][x+1] = (byte) 6;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(contextMap[y+1][x]){
				case 0:
					contextMap[y+1][x] = (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y+1][x] = (byte) 3;
					break;
				case 3:
					contextMap[y+1][x] = (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y+1][x] = (byte) 7;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(contextMap[y-1][x]){
				case 0:
					contextMap[y-1][x] = (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y-1][x] = (byte) 3;
					break;
				case 3:
					contextMap[y-1][x] = (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y-1][x] = (byte) 7;
					break;
				}
			}
			break;
		case 1: //SUBBAND HL
			if(x-1 >= 0){
				//LEFT
				switch(contextMap[y][x-1]){
				case 0:
					contextMap[y][x-1] = (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y][x-1] = (byte) 3;
					break;
				case 3:
					contextMap[y][x-1] = (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y][x-1] = (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(contextMap[y-1][x-1]){
					case 0:
						contextMap[y-1][x-1] = (byte) 1;
						break;
					case 1:
						contextMap[y-1][x-1] = (byte) 2;
						break;
					case 5:
						contextMap[y-1][x-1] = (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(contextMap[y+1][x-1]){
					case 0:
						contextMap[y+1][x-1] = (byte) 1;
						break;
					case 1:
						contextMap[y+1][x-1] = (byte) 2;
						break;
					case 5:
						contextMap[y+1][x-1] = (byte) 6;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(contextMap[y][x+1]){
				case 0:
					contextMap[y][x+1] = (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y][x+1] = (byte) 3;
					break;
				case 3:
					contextMap[y][x+1] = (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y][x+1] = (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(contextMap[y-1][x+1]){
					case 0:
						contextMap[y-1][x+1] = (byte) 1;
						break;
					case 1:
						contextMap[y-1][x+1] = (byte) 2;
						break;
					case 5:
						contextMap[y-1][x+1] = (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(contextMap[y+1][x+1]){
					case 0:
						contextMap[y+1][x+1] = (byte) 1;
						break;
					case 1:
						contextMap[y+1][x+1] = (byte) 2;
						break;
					case 5:
						contextMap[y+1][x+1] = (byte) 6;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(contextMap[y+1][x]){
				case 0:
					contextMap[y+1][x] = (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y+1][x] = (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y+1][x] = (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y+1][x] = (byte) 8;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(contextMap[y-1][x]){
				case 0:
					contextMap[y-1][x] = (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y-1][x] = (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y-1][x] = (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y-1][x] = (byte) 8;
					break;
				}
			}
			break;
		case 3: //SUBBAND HH
			if(x-1 >= 0){
				//LEFT
				switch(contextMap[y][x-1]){
				case 0:
					contextMap[y][x-1] = (byte) 1;
					break;
				case 1:
					contextMap[y][x-1] = (byte) 2;
					break;
				case 3:
					contextMap[y][x-1] = (byte) 4;
					break;
				case 4:
					contextMap[y][x-1] = (byte) 5;
					break;
				case 6:
					contextMap[y][x-1] = (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(contextMap[y-1][x-1]){
					case 0:
						contextMap[y-1][x-1] = (byte) 3;
						break;
					case 1:
						contextMap[y-1][x-1] = (byte) 4;
						break;
					case 2:
						contextMap[y-1][x-1] = (byte) 5;
						break;
					case 3:
						contextMap[y-1][x-1] = (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y-1][x-1] = (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y-1][x-1] = (byte) 8;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(contextMap[y+1][x-1]){
					case 0:
						contextMap[y+1][x-1] = (byte) 3;
						break;
					case 1:
						contextMap[y+1][x-1] = (byte) 4;
						break;
					case 2:
						contextMap[y+1][x-1] = (byte) 5;
						break;
					case 3:
						contextMap[y+1][x-1] = (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y+1][x-1] = (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y+1][x-1] = (byte) 8;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(contextMap[y][x+1]){
				case 0:
					contextMap[y][x+1] = (byte) 1;
					break;
				case 1:
					contextMap[y][x+1] = (byte) 2;
					break;
				case 3:
					contextMap[y][x+1] = (byte) 4;
					break;
				case 4:
					contextMap[y][x+1] = (byte) 5;
					break;
				case 6:
					contextMap[y][x+1] = (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(contextMap[y-1][x+1]){
					case 0:
						contextMap[y-1][x+1] = (byte) 3;
						break;
					case 1:
						contextMap[y-1][x+1] = (byte) 4;
						break;
					case 2:
						contextMap[y-1][x+1] = (byte) 5;
						break;
					case 3:
						contextMap[y-1][x+1] = (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y-1][x+1] = (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y-1][x+1] = (byte) 8;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(contextMap[y+1][x+1]){
					case 0:
						contextMap[y+1][x+1] = (byte) 3;
						break;
					case 1:
						contextMap[y+1][x+1] = (byte) 4;
						break;
					case 2:
						contextMap[y+1][x+1] = (byte) 5;
					break;
					case 3:
						contextMap[y+1][x+1] = (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y+1][x+1] = (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y+1][x+1] = (byte) 8;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(contextMap[y+1][x]){
				case 0:
					contextMap[y+1][x] = (byte) 1;
					break;
				case 1:
					contextMap[y+1][x] = (byte) 2;
					break;
				case 3:
					contextMap[y+1][x] = (byte) 4;
					break;
				case 4:
					contextMap[y+1][x] = (byte) 5;
					break;
				case 6:
					contextMap[y+1][x] = (byte) 7;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(contextMap[y-1][x]){
				case 0:
					contextMap[y-1][x] = (byte) 1;
					break;
				case 1:
					contextMap[y-1][x] = (byte) 2;
					break;
				case 3:
					contextMap[y-1][x] = (byte) 4;
					break;
				case 4:
					contextMap[y-1][x] = (byte) 5;
					break;
				case 6:
					contextMap[y-1][x] = (byte) 7;
					break;
				}
			}
			break;
		}
	}
}
