/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOIDecode;
import BOIException.ErrorException;
import BOIException.ParameterException;
import BOIParser.ArgumentsParser;

import java.lang.reflect.Method;


/**
 * Arguments parser for BOI decoder (extended from ArgumentsParser).
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class DecoderParser extends ArgumentsParser{

	//ARGUMENTS SPECIFICATION
	String[][] decoderArguments = {
		{"-i", "--inputFile", "{string}", "", "1", "1",
			"Input JPEG2000 file."
		},
		{"-o", "--outputImage", "{string}", "", "1", "0",
			"Output image file name where recovered image samples will be stored. Image file type will be decided depending on the extension. Valid file types are:\n    RAW/IMG raw data (if outputImage is raw data, \"-og\" argument is mandatory)\n    PNM/PGM/PPM PGM is used when image have 1 component, PPM when there are 3 components and PNM for both\n    TIFF\n    JPEG (not recommended because it degenerates recovered image)\n    BMP (not recommended)\n    ATTENTION: if recovered image samples are signed only RAW/IMG data with a signed data type will preserve image samples consistently. Using other file format will cause image data damages."
		},
		{"-og", "--ouputImageGeometry", "{int int}", "", "0", "1",
			"Image raw data type. Parameters are:\n    1- data type. Possible values are:\n      0- boolean (1 byte)\n      1- unsigned int (1 byte)\n      2- unsigned int (2 bytes)\n      3- signed int (2 bytes)\n      4- signed int (4 bytes)\n      5- signed int (8 bytes)\n      6- float (4 bytes)\n      7- double (8 bytes)\n    2- Byte order (0 if BIG ENDIAN, 1 if LITTLE ENDIAN)"
		},
		{"-vc", "--verboseComputation", "{boolean}", (DecoderDefaultValues.CVerbose[0] ? "1": "0") + " ", "0", "1",
			"Show some information about time and used memory for each compression stage. Value is a boolean: 0 indicates NO show and 1 indicates show."
		},
		{"-ve", "--verboseEncodingParameters", "{boolean}", (DecoderDefaultValues.CVerbose[1] ? "1": "0") + " ", "0", "1",
			"Show BOI decoding parameters (taken from the file heading). Value is a boolean: 0 indicates NO show and 1 indicates show."
		},
		{"-tn", "--threadNumber", "{boolean}", DecoderDefaultValues.CThreadNumber + " ", "0", "1",
			"Number of threads employed in the block coding stage."
		},
		{"-h", "--help", "", "", "0", "1",
			"Displays this help and exits program."
		}
	};

	//ARGUMENTS VARIABLES
	String DInFile = "";
	String[] imageFiles = null;
	int[] imageGeometry = null;
	int CThreadNumber = -1;
	boolean[] CVerbose = null;
	//Number of imageFiles specified
	int ifs = 0;


	/**
	 * Receives program arguments and parses it, setting to arguments variables.
	 *
	 * @param arguments the array of strings passed at the command line
	 *
	 * @throws ParameterException when an invalid parsing is detected
	 * @throws ErrorException when some problem with method invocation occurs
	 */
	public DecoderParser(String[] arguments) throws ParameterException, ErrorException{
		try{
			Method m = this.getClass().getMethod("parseArgument", new Class[] {int.class, String[].class});
			parse(decoderArguments, arguments, this, m);
		}catch(NoSuchMethodException e){
			throw new ErrorException("Decoder parser error invoking parse function.");
		}
	}

	/**
	 * Parse an argument using parse functions from super class and put its value/s to the desired variable. This function is called from parse function of the super class.
	 *
	 * @param argFound number of parameter (the index of the array coderArguments)
	 * @param options the command line options of the argument
	 *
	 * @throws ParameterException when some error about parameters passed (type, number of params, etc.) occurs
	 */
	public void parseArgument(int argFound, String[] options) throws ParameterException{
		switch(argFound){
		case  0: //-i  --inputFile
			DInFile = parseString(options);
			break;
		case  1: //-o  --outputImage
			if(imageFiles != null){
				String[] imageFilesTMP = new String[imageFiles.length + 1];
				for(int file = 0; file < imageFiles.length; file++){
					imageFilesTMP[file] = imageFiles[file];
				}
				imageFiles = imageFilesTMP;
			}else{
				imageFiles = new String[1];
			}
			imageFiles[ifs] = parseString(options);
			if(imageFiles[ifs].endsWith(".raw") || imageFiles[ifs].endsWith(".img")){
				decoderArguments[2][4] = "1";
			}
			ifs++;
			break;
		case  2: //-og  --outputImageGeometry
			imageGeometry = parseIntegerArray(options, 2);
			break;
		case  3: //-vc  --verboseComputation
			if(CVerbose == null){
				CVerbose = DecoderDefaultValues.CVerbose;
			}
			CVerbose[0] = parseBoolean(options);
			break;
		case  4: //-vc  --verboseParameters
			if(CVerbose == null){
				CVerbose = DecoderDefaultValues.CVerbose;
			}
			CVerbose[1] = parseBoolean(options);
			break;
		case  5: //-tn  --threadNumber
			CThreadNumber = parseIntegerPositive(options);
			break;
		case  6: //-h  --help
			System.out.println("BOI Decoder 1.6");
			showArgsInfo();
			//showArgsInfoLatexTable();
			System.exit(0);
			break;
		}
	}

	//ARGUMENTS GET FUNCTIONS
	public String getDInFile(){
		return(DInFile);
	}
	public String[] getImageFiles(){
		return(imageFiles);
	}
	public int[] getImageGeometry(){
		return(imageGeometry);
	}
	public boolean[] getCVerbose(){
		return(CVerbose);
	}
	public int getCThreadNumber(){
		return(CThreadNumber);
	}

}

